package ICQServer;
#
# Perl interface for the ICQ server.
#
# This program was made without any help from Mirabilis or their
# consent.  No reverse engineering or decompilation of any Mirabilis
# code took place to make this program.
#
# Copyright (c) 1999 Peter Kjellerstedt and Fredrik Norrman.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the same terms as Perl itself (GPL).
#
#
#  HISTORY
#
#  DATE         NAME               CHANGES
#  ----         ----               -------
#  Apr 10 1999  Fredrik & Peter    Initial version
#  Apr 14 1999  Fredrik & Peter    Implemented the version 5 encryption
#  Apr 24 1999  Fredrik Norrman    Rewrote the database to use SQL
#  May 30 1999  Fredrik & Peter    Fixed most of the information passing
#                                  between server and client
#  Oct  4 1999  Peter Kjellerstedt Prepared for public distribution
#
#
# TODO:
# * Set security reply needs something extra in the reply when successful
# * Handle messages sent to an unknown socket gracefully...
# * Resend packets if no ack is returned
# * Should probably send SILENT_TOO_LONG a couple of times before kicking
#   off someone
# * What are the last two zeroes sent in construct_meta_info_main?
# * Whitepages search (backgrounds, interests, and organizations)
# * Still a few unknowns in the command tables, of which some will hang
#   the server/client connection if received...
# * Doesn't handle info requests for unknown UINs correctly
# * More documentation... ;-)
#
# DELIMITATION
# * This server only support version 5 of the ICQ protocol. It originally
#   supported version 2, but that support was lost along the way.
#

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK 
            %status_bits %user_status 
            %client_commands_bynumber 
            %server_commands_bynumber 
            %client_meta_commands_bynumber 
            %server_meta_commands_bynumber);

require Exporter;

use Data::Dumper;
use IO::Socket;
use IO::Select;
use Sys::Hostname;
use DBI;
use Symbol;
use Fcntl;

use config;

use Carp;                      # Regular Carp
#use Carp qw(verbose);         # This one's for debugging - uses line numbers

@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw();

$VERSION = "0.04";



=head1 NAME

ICQServer - An ICQ server

=head1 SYNOPSIS

   use ICQServer;
     
   $ICQServer = ICQServer->new();
   while ($ICQServer->incoming_packet_waiting())
   {
      $ICQServer->incoming_process_packet();
   }

=head1 DESCRIPTION

C<ICQServer> is a class implementing a simple ICQ server in
Perl.

=cut

###################################################################
# Some constants                                                  #
###################################################################

my $DEBUG = 1;
my $DEBUG_ACK = 0;
my $DEBUG_PACKET_DUMP = 1;

# We can handle up to ICQ Ver 5.0
my $icq_max_version = 5;

my $session_timeout = 60*2+30; # seconds
my $max_search_hits = 40;

# Status types
# TODO there are new status options that I don't have
%status_bits =
(
  "ONLINE"        => 0x00000000,
  "AWAY"          => 0x00000001,
  "ACCEPT_URGENT" => 0x00000002, # ?????
  "NA"            => 0x00000004,
  "DND"           => 0x00000010,
  "FREE_CHAT"     => 0x00000020,
  "INVISIBLE"     => 0x00000100,
  "UNKNOWN"       => 0x00010000,
  "BIRTHDAY"      => 0x00080000,
  "OFFLINE"       => 0xFFFFFFFF
);

# 00001  Away
# 00002  
# 00004  NA
# 00010  DND
# 00020  Free Chat
# 00100  Invisible
# 10000  
# 80000  Birthday



# Packet types
%server_commands_bynumber =
(
  0x000A => "ACK",
  0x0028 => "SILENT_TOO_LONG",
  0x0032 => "OLD_USER_UIN",
  0x003C => "OLD_USER_UIN_V5", # only bad login?
  0x0046 => "NEW_USER_UIN",
  0x005A => "LOGIN_REPLY",
  0x0064 => "BAD_LOGIN",
  0x006E => "USER_ONLINE",
  0x0078 => "USER_OFFLINE",
  0x0082 => "QUERY_REPLY",
  0x008C => "USER_FOUND",
  0x00A0 => "END_OF_SEARCH",
  0x00B4 => "NEW_USER_REPLY",
  0x00C8 => "UPDATE_EXT_REPLY",
  0x00DC => "RECEIVE_MESSAGE",
  0x00E6 => "END_OFFLINE_MESSAGES",
  0x00F0 => "NOT_LOGGED_IN",
  0x00FA => "TRY_AGAIN",
  0x0104 => "SYSTEM_DELIVERED_MESSAGE",
  0x0118 => "INFO_REPLY",
  0x0122 => "EXT_INFO_REPLY",
  0x01A4 => "STATUS_UPDATE",
  0x01C2 => "SYSTEM_MESSAGE",
  0x01CC => "DELETE_SUCCEEDED",
  0x01D6 => "DELETE_FAILED",
  0x01E0 => "UPDATE_SUCCEEDED",
  0x01EA => "UPDATE_FAILED",
  0x01F4 => "AUTH_REPLY",
  0x0212 => "MULTI_PACKET",
  0x021C => "END_CONTACTLIST_STATUS",
  0x024E => "RANDOM_USER_REPLY",
  0x03DE => "META_USER_REPLY",
);
my %server_commands = reverse %server_commands_bynumber;

%client_commands_bynumber =
(
  0x000A => "ACK",
  0x010E => "SEND_MESSAGE",
  0x03E8 => "LOGIN",
  0x03F2 => "OLD_USER_REG",
  0x03FC => "NEW_USER_REG",
  0x0406 => "CONTACT_LIST",
  0x041A => "SEARCH_UIN",
  0x0424 => "SEARCH_USER",
  0x042E => "KEEP_ALIVE",
  0x0438 => "SEND_TEXT_CODE",
  0x0442 => "ACK_OFFLINE_MESSAGES",
  0x044C => "REQUEST_OFFLINE_MESSAGES",
  0x0456 => "REQ_ADD_TO_LIST",
  0x0460 => "INFO_REQ",
  0x046A => "EXT_INFO_REQ",
  0x049C => "CHANGE_PASSWORD",
  0x04A6 => "NEW_USER_INFO",
  0x04B0 => "UPDATE_EXT_INFO",
  0x04BA => "QUERY_SERVERS",
  0x04C4 => "QUERY_ADDONS",
  0x04D8 => "STATUS_CHANGE",
  0x04EC => "NEW_USER_1",
  0x0500 => "DELETE_UIN",
  0x050A => "UPDATE_INFO",
  0x0514 => "AUTH_UPDATE",
  0x051E => "KEEP_ALIVE_2",
  0x0528 => "LOGIN_2",
  0x053C => "ADD_TO_LIST",
  0x0564 => "RANDOM_SET",
  0x056E => "RANDOM_SEARCH",
  0x064A => "META_USER",
  0x06A4 => "INVISIBLE_LIST",
  0x06AE => "VISIBLE_LIST",
  0x06B8 => "VISIBILITY_LISTS",

  0x04CE => "CLIENT_UNKNOWN_1",
  0x04F6 => "CLIENT_UNKNOWN_2",
  0x0532 => "CLIENT_UNKNOWN_3",
);
my %client_commands = reverse %client_commands_bynumber;

%client_meta_commands_bynumber =
(
#  0x00C8 => "META_SRV_GEN",
#  0x00D2 => "META_SRV_WORK",
#  0x00DC => "META_SRV_MORE",
#  0x00AA => "META_SRV_PASS",
  0x03E8 => "META_INFO_SET_MAIN",
  0x03F2 => "META_INFO_SET_WORK",
  0x03FC => "META_INFO_SET_MORE",
  0x0406 => "META_INFO_SET_ABOUT",
  0x0410 => "META_INFO_SET_INTERESTS",
  0x041A => "META_INFO_SET_AFFILIATIONS",
  0x0424 => "META_INFO_SET_SECURITY",
  0x042E => "META_INFO_SET_PASSWORD",
  0x0442 => "META_INFO_SET_HOMEPAGE_CATEGORY",
  0x04B0 => "META_INFO_GET_UNKNOWN1",
  0x04BA => "META_INFO_UNKNOWN1",
  0x04CE => "META_INFO_ALL",
  0x0514 => "META_INFO_SEARCH_DETAILS",
  0x051E => "META_INFO_SEARCH_UIN",
  0x0528 => "META_INFO_SEARCH_EMAIL",
  0x0532 => "META_INFO_SEARCH_WHITEPAGES",
  0x053C => "META_INFO_SEARCH_DETAILS_WILDCARD",
  0x0546 => "META_INFO_SEARCH_EMAIL_WILDCARD",
  0x0550 => "META_INFO_SEARCH_WHITEPAGES_WILDCARD",
  0x06B8 => "META_INFO_UNKNOWN2",
);
my %client_meta_commands = reverse %client_meta_commands_bynumber;


%server_meta_commands_bynumber =
(
  0x0064 => "META_SRV_SET_MAIN_REPLY",
  0x006E => "META_SRV_SET_WORK_REPLY",
  0x0078 => "META_SRV_SET_MORE_REPLY",
  0x0082 => "META_SRV_SET_ABOUT_REPLY",
  0x008C => "META_SRV_SET_INTERESTS_REPLY",
  0x0096 => "META_SRV_SET_AFFILIATIONS_REPLY",
  0x00A0 => "META_SRV_SET_SECURITY_REPLY",
  0x00AA => "META_SRV_SET_PASSWORD_REPLY",
  0x00BE => "META_SRV_SET_HOMEPAGE_CATEGORY_REPLY",
  0x00C8 => "META_SRV_INFO_MAIN",
  0x00D2 => "META_SRV_INFO_WORK",
  0x00DC => "META_SRV_INFO_MORE",
  0x00E6 => "META_SRV_INFO_ABOUT",
  0x00F0 => "META_SRV_INFO_INTERESTS",
  0x00FA => "META_SRV_INFO_AFFILIATIONS",
  0x010E => "META_SRV_INFO_HOMEPAGE_CATEGORY",
  0x0190 => "META_SRV_SEARCH_REPLY",
  0x019A => "META_SRV_SEARCH_LAST_REPLY",
  0x01A4 => "META_SRV_SEARCH_WHITEPAGES_REPLY",
  0x01AE => "META_SRV_SEARCH_WHITEPAGES_LAST_REPLY",
  0x02D0 => "META_SRV_UNKNOWN_2",
);
my %server_meta_commands = reverse %server_meta_commands_bynumber;


# 0064 0A  ###  set
# 006E 0A  ###  work
# 0078 0A  ###  more
# 0082 0A  ###  about
#   8C     #    interests
#   96 0A  #    affiliations
# 00A0 0A  ###  secure
#   AA     #    password
#   B4
# 00BE 0A  ###  homepage_category
# 00D0
# 02D0 14  ###  unknown1

# 00C8 nick, address
# 00D2
# 00DC
# 00E6 about?
# 010E

my(@month_days) = (31, 28, 30, 31, 30, 31, 31, 30, 31, 30, 31);

###################################################################
# Functions under here are member functions                       #
###################################################################

=head1 CONSTRUCTOR

=item new ( [ PORT ])

Establish the ICQ server and await connections.

C<PORT> is the port to which the server will respond.

The constructor returns the open socket, or C<undef> if an error has
been encountered.

=cut

sub new
{
  my $prototype = shift;
  my $port      = shift;

  my $class = ref($prototype) || $prototype;
  my $self  = {};

  $self->{'port'} = $port || $ENV{'ICQPORT'} || $config::default_icq_port;
  chomp($self->{'tty'} = `tty`);
   
  $DEBUG && warn "ICQServer port: " . $self->{'port'} . "\n";

  # open the connection
  $self->{'socket'} =
    new IO::Socket::INET(LocalPort => $self->{'port'},
                         Proto     => "udp",
                         Type      => SOCK_DGRAM,
                         Reuse     => 1
                         ) || croak "new: connect socket: $!";

  $self->{'select'} = new IO::Select [$self->{'socket'}];

  bless($self, $class);

  $self->db_init;
  
  return $self;
}

#
# destructor
#
sub DESTROY
{
  my $self = shift;

  $self->db_close;
  shutdown($self->{'socket'}, 2);
  close($self->{'socket'});

  return 1;
}


=head1 INCOMING - HIGH LEVEL FUNCTIONS

Copes with responses from the ICQ server.

=item incoming_packet_waiting ( TIMEOUT );

Check if there is something from the server waiting to be processed.

To have it block waiting for input, call it with no argument.
Otherwise the argument is the number of seconds before it times out.

=cut

sub incoming_packet_waiting
{
  my $self    = shift;
  my $timeout = shift;

  return $self->{'select'}->can_read($timeout);
}

=pod

=item incoming_process_packet ( );

Read a packet from the socket. Then dispatch a suitable method that
will respond to the packet.

=cut

sub incoming_process_packet
{
  my $self = shift;

  my $server;
  my $packet;
   
#  $DEBUG && warn "INCOMING:  Reading incoming packet ...\n";

  if (!defined($self->{'socket'}))
  {
    die "No socket!\n";
    return;
  }

  my $sock = $self->{'socket'};
  $self->{'user_uin'} = 0;

  unless ($sock->recv($packet, 99999))
  {
#   croak "socket:  recv2:  $!";

    print "Socket died! Opening a new...\n";

    # open the connection
    $self->{'socket'} =
      new IO::Socket::INET(LocalPort => $self->{'port'},
                           Proto     => "udp",
                           Type      => SOCK_DGRAM,
                           Reuse     => 1
                           ) || croak "new: connect socket: $!";

    $self->{'select'} = new IO::Select [$self->{'socket'}];
    return;
  }

  if (!defined($packet))
  {
    die "No message!\n";
    return;
  }

  my($version, $command, $sequence_number, $session_id, $zeroes, 
     $sequence_number2, $user_uin, $checksum);

  $version = unpack("v", $packet);
  if ($version == 2)
  {
    if ($DEBUG && $DEBUG_PACKET_DUMP)
    {
      print "\nIncoming packet:\n";
      $self->dump_packet($packet);
    }
    
    ($packet, $version)         = extract_uword($packet);
    ($packet, $command)         = extract_uword($packet);
    ($packet, $sequence_number) = extract_uword($packet);

    ($zeroes, $sequence_number2, $user_uin, $checksum, $session_id) = 
      (0, 0, 0, 0, 0);
  }
  elsif ($version == 3 || $version == 4)
  {
#    if ($DEBUG && $DEBUG_PACKET_DUMP)
#    {
#      print "\nIncoming encrypted packet:\n";
#      $DEBUG_PACKET_DUMP && $self->dump_packet($packet);
#    }

    $packet = $self->decrypt_packet_v4($packet);

    ($packet, $version)          = extract_uword($packet);
    ($packet, $session_id)       = extract_uword($packet);
    ($packet, $zeroes)           = extract_uword($packet);
    ($packet, $command)          = extract_uword($packet);
    ($packet, $sequence_number)  = extract_uword($packet);
    ($packet, $sequence_number2) = extract_uword($packet);
    ($packet, $user_uin)         = extract_udword($packet);
    ($packet, $checksum)         = extract_udword($packet);
  }
  elsif ($version == 5)
  {
#    if ($DEBUG && $DEBUG_PACKET_DUMP)
#    {
#      print "\nIncoming encrypted packet:\n";
#      $self->dump_packet($packet);
#    }
    
    my($checkcode) = $self->find_checkcode_v5($packet);
    $packet = $self->crypt_packet_v5($packet, $checkcode);

    my($org_packet) = $packet;
    
    ($packet, $version)          = extract_uword($packet);
    ($packet, $zeroes)           = extract_udword($packet);
    ($packet, $user_uin)         = extract_udword($packet);
    ($packet, $session_id)       = extract_udword($packet);
    ($packet, $command)          = extract_uword($packet);
    ($packet, $sequence_number)  = extract_uword($packet);
    ($packet, $sequence_number2) = extract_uword($packet);
    ($packet, $checksum)         = extract_udword($packet);

    if ($DEBUG && $DEBUG_PACKET_DUMP &&
        ($DEBUG_ACK ||
          ($command ne $client_commands{'ACK'} &&
           $command ne $client_commands{'KEEP_ALIVE'} &&
           $command ne $client_commands{'KEEP_ALIVE_2'})))
    {
      print "\nIncoming decrypted packet:\n";
      $self->dump_packet($org_packet);
    }
    
  }
  else
  {
    $DEBUG && warn("Cannot handle version $version packets!!!\n");
    return;
  }
  
  if ($DEBUG)
  {
    my $command_hex = &uword_to_hex($command);
    my $sc = $client_commands_bynumber{$command};
    if ($sc)
    {
      if ($DEBUG_ACK ||
          ($command ne $client_commands{'ACK'} &&
           $command ne $client_commands{'KEEP_ALIVE'} &&
           $command ne $client_commands{'KEEP_ALIVE_2'}))
      {
        warn "INCOMING:  Got command " .  
          $client_commands_bynumber{$command} . " ($command/$command_hex) ",
          "sequence $sequence_number, $sequence_number2\n";
      }
    }
    else
    {
      warn "INCOMING:  Got unknown command number $command/$command_hex ",
      "sequence $sequence_number, $sequence_number2\n";
    }
  }

  if ($command eq $client_commands{'ACK'}) {
    # ack is special case, we ignore it for the moment
    # TODO we should keep track of packets sent (hash, indexed
    #      by sequence number) and tick them off when they're
    #      ACK'd.  Then resend things if they aren't.  Fiddly.
    $DEBUG && $DEBUG_ACK && 
      warn "INCOMING:  Got ACK for packet $sequence_number\n";
  }
  else
  {
    # common packet info
    my(%packet_header) =
    (
      'version'    => $version,
      'command'    => $command,
      'seq_num'    => $sequence_number,
      'seq_num2'   => $sequence_number2,
      'user_uin'   => $user_uin,
      'checksum'   => $checksum,
      'session_id' => $session_id,
    );

    $self->{'packet_header'} = \%packet_header;
    $DEBUG && $DEBUG_ACK && 
      warn "INCOMING: Sending ACK for packet $sequence_number, ",
      "$sequence_number2, random=$session_id\n";
    
    # If this is a disconnect message we must not ack it, so the ack is
    # done in receive_send_text_code if needed...
    unless ($command == $client_commands{'SEND_TEXT_CODE'})
    {
      $self->send_ack($sequence_number, $sequence_number2);
    }

    if ($self->session_check_user_status($user_uin, 'OFFLINE'))
    {
      # User is set offline => we only allow session setup commands 
      if ($command != $client_commands{'LOGIN'} &&
          $command != $client_commands{'OLD_USER_REG'} &&
          $command != $client_commands{'NEW_USER_REG'} &&
          $command != $client_commands{'NEW_USER_1'} &&
          $command != $client_commands{'SEND_TEXT_CODE'})
      {
        print(">>NOT_LOGGED_IN:\n",
              "    user_uin=$user_uin\n");

        $packet = "";

        $self->send_packet($self->construct_packet("NOT_LOGGED_IN", $packet));
        return 0;
      }
    }

    # Update time stamp
    if ($user_uin != 0 && $command != $client_commands{'LOGIN'})
    {
      $self->{'sessions'}{$user_uin}{'last_packet_received_time'} = time;
    }
    
    $self->session_kill_inactive_sessions();
  
    my $command_name = "receive_" . lc($client_commands_bynumber{$command});
    my $coderef = $self->can($command_name);
    if ($command_name && $coderef)
    {
      # we've found a method that can process this type of packet
      return $self->$command_name($packet);
    }
    else
    {
      # TODO: can't cope - what do we do with this packet!!
      warn("UNKNOWN PACKET TYPE: '$command/",
           &uword_to_hex($command),
           "', sequence number '$sequence_number, $sequence_number2'");
      $DEBUG && warn("INCOMING: Unknown packet dump: ", Dumper($packet), "\n");
      return 0;
    }
  }
}


=head1 OUTGOING - LOW LEVEL FUNCTIONS

These correspond directly with the packets available in the ICQ
protocol.

=item send_ack ( SEQUENCE_NUMBER );

Send an ACK to the server, confirming we got packet SEQUENCE_NUMBER.

=cut

sub send_ack
{
  my $self     = shift;
  my $seq_num  = shift;
  my $seq_num2 = shift;

  $DEBUG && $DEBUG_ACK && warn ">>ACK\n";
  return $self->send_packet($self->construct_packet("ACK", "",
                                                    $seq_num, $seq_num2));
}

sub send_multi_packet
{
  my $self     = shift;
  my $packets  = $#_ + 1;
  my $user_uin = $self->{'packet_header'}->{'user_uin'};
  my $seq      = $self->{'sessions'}{$user_uin}{'sequence_number'} - $packets;

  if ($packets == 1)
  {
    $self->send_packet($_[0]);
    return;
  }
  
  my $packet = "";
  $packet = insert_byte($packet, $packets);

  foreach (@_)
  {
    $packet = insert_uword($packet, length($_));  
    $packet .= $_;
  }
    
  $self->send_packet($self->construct_packet("MULTI_PACKET", $packet, $seq,0));
}

sub send_meta_info_search_reply
{
  my $self          = shift;
  my $is_whitepages = shift;
  my $reply_code;
  my $last_reply_code;
  
  if ($is_whitepages)
  {
    $reply_code      = 'META_SRV_SEARCH_WHITEPAGES_REPLY';
    $last_reply_code = 'META_SRV_SEARCH_WHITEPAGES_LAST_REPLY';
  }
  else
  {
    $reply_code      = 'META_SRV_SEARCH_REPLY';
    $last_reply_code = 'META_SRV_SEARCH_LAST_REPLY';
  }

  my($found) = $#_ + 1;

  if (!$found)
  {
    my $packet = "";
    $packet = insert_uword($packet, $server_meta_commands{$last_reply_code});
    $packet = insert_byte ($packet, 0x32);

    $self->send_packet($self->construct_packet("META_USER_REPLY", $packet));
    return;
  }

  my(@packets) = ();

  my($sent_uins) = 0;
  foreach my $searched_uin (@_)
  {
    my($user_info) = $self->db_user_info($searched_uin);
  
    if (defined($user_info))
    {
      my($nick_name, $first_name, $last_name, $email, $authorization_status) =
        ($$user_info{'nick_name'},
         $$user_info{'first_name'},
         $$user_info{'last_name'},
         $$user_info{'email'},
         $$user_info{'authorization'});

      my $last_uin = (++$sent_uins == $max_search_hits ||
                      $sent_uins == $found);

      my($command) = ($last_uin ?
		      $server_meta_commands{$last_reply_code} :
		      $server_meta_commands{$reply_code});

      my $status;
      if ($self->session_check_user_status($searched_uin, 'OFFLINE'))
      {
	$status = 0;
      }
      elsif ($self->session_check_user_status($searched_uin, 'ONLINE'))
      {
	$status = 1;
      }
      else
      {
	$status = 2;
      }

      my $packet = "";
      $packet = insert_uword ($packet, $command);
      $packet = insert_byte  ($packet, 0x0A);

      $packet = insert_udword($packet, $searched_uin);
      $packet = insert_string($packet, $nick_name);
      $packet = insert_string($packet, $first_name);
      $packet = insert_string($packet, $last_name);
      $packet = insert_string($packet, $email);
      $packet = insert_byte  ($packet, $authorization_status);
      $packet = insert_uword ($packet, $status);

      if ($last_uin)
      {
        $packet = insert_udword($packet, max(0, $found - $max_search_hits));
      }
      
      push(@packets, $self->construct_packet("META_USER_REPLY", $packet));
      
      last if ($last_uin);
    }
  }
  
  $self->send_multi_packet(@packets);
}

=head1 INCOMING - LOW LEVEL FUNCTIONS

Copes with requests from any ICQ client at the packet level.

=item receive_login ( PACKET );

Receive the login packet from the ICQ socket and respond to it appropriately.

=cut

sub receive_login
{
  my $self   = shift;
  my $packet = shift;

  my $packet_header = $self->{'packet_header'};
  
  print "Using login version: " . $packet_header->{'version'} . "\n";

  if ($packet_header->{'version'} == 2)
  {  
    my($user_uin, $tcp_port, $password, $unknown1, $ip_address, $unknown2,
       $status, $unknown3, $login_seq_num, $unknown4, $unknown5);
    ($packet, $user_uin)      = extract_udword($packet); 
    ($packet, $tcp_port)      = extract_udword($packet); 
    ($packet, $password)      = extract_string($packet);
    ($packet, $unknown1)      = extract_udword($packet); 
    ($packet, $ip_address)    = extract_ipaddr($packet); 
    ($packet, $unknown2)      = extract_byte($packet); 
    ($packet, $status)        = extract_udword($packet); 
    ($packet, $unknown3)      = extract_udword($packet); 
    ($packet, $login_seq_num) = extract_uword($packet); 
    ($packet, $unknown4)      = extract_udword($packet); 
    ($packet, $unknown5)      = extract_udword($packet); 


    $DEBUG && warn("<<LOGIN:\n",
                   "    user_uin=$user_uin\n",
                   "    tcp_port=$tcp_port\n",
                   "    password=$password\n",
                   "    ip_address=", inet_ntoa(pack("N", $ip_address)), "\n",
                   "    status=", $self->status_to_text($status), "\n",
                   "    login_seq_num=$login_seq_num\n",
                   "    unknown1=", udword_to_hex($unknown1), "\n",
                   "    unknown2=$unknown2\n",
                   "    unknown3=", udword_to_hex($unknown3), "\n",
                   "    unknown4=", udword_to_hex($unknown4), "\n",
                   "    unknown5=", udword_to_hex($unknown5), "\n");
    
    if (!$self->verify_password($user_uin, $password))
    {
      $self->send_packet($self->construct_packet("BAD_LOGIN",
                                                 pack("vV",
                                                      $login_seq_num,
                                                      $user_uin)));
      return;
    }
    
    if (!$self->session_check_user_status($user_uin, 'OFFLINE'))
    {
      $self->send_packet($self->construct_packet("TRY_AGAIN",
                                                 pack("vV2",
                                                      0x006E, $user_uin, 0)));
      # Just a guess, but the 0x006E above looks very much like the code for
      # the server command USER_ONLINE...
    }
    else
    {
      $self->{'user_uin'} = $user_uin;
      
      $self->{'sessions'}{$user_uin}{'sequence_number'} = 0;
      $self->{'sessions'}{$user_uin}{'ip_address'} = $self->{'socket'}->peerhost;
      $self->{'sessions'}{$user_uin}{'udp_port'} = $self->{'socket'}->peerport;
      $self->{'sessions'}{$user_uin}{'real_ip_address'} = $ip_address;
      $self->{'sessions'}{$user_uin}{'tcp_port'} = $tcp_port;
      $self->{'sessions'}{$user_uin}{'last_packet_received_time'} = time;
      
      $self->send_packet($self->construct_packet("LOGIN_REPLY",
                                                     pack("VNvN7 ",
                                                          $user_uin,
                                                          $ip_address,
                                                          $login_seq_num,
                                                          0x01000100,
                                                          0x17001600,
                                                          0x8C000000,
                                                          0xF0000A00,
                                                          0x0A000500,
                                                          0x0A0001AC,
                                                          0xC2AC1201)));
      
      $self->session_set_user_status($user_uin, $status);
    }
  }
  elsif ($packet_header->{'version'} == 5)
  {
    my $user_uin = $packet_header->{'user_uin'};
    my($time, $tcp_port, $password, $unknown1, $ip_address, $unknown2,
       $status, $unknown3, $login_seq_num, $unknown4, $unknown5,
       $unknown6, $unknown7, $unknown8);
    ($packet, $time)          = extract_udword($packet);
    ($packet, $tcp_port)      = extract_udword($packet);
    ($packet, $password)      = extract_string($packet);
    ($packet, $unknown1)      = extract_udword($packet);
    ($packet, $ip_address)    = extract_ipaddr($packet);
    ($packet, $unknown2)      = extract_byte($packet);
    ($packet, $status)        = extract_udword($packet);
    ($packet, $unknown3)      = extract_udword($packet);
    ($packet, $unknown4)      = extract_udword($packet);
#   ($packet, $unknown5)      = extract_udword($packet);
#   ($packet, $unknown6)      = extract_udword($packet);
#   ($packet, $unknown7)      = extract_udword($packet);
#   ($packet, $unknown8)      = extract_udword($packet);
    
    $DEBUG && warn("<<LOGIN:\n",
                   "    user_uin=$user_uin\n",
                   "    time=$time\n",
                   "    tcp_port=$tcp_port\n",
                   "    password=$password\n",
                   "    ip_address=", inet_ntoa(pack("N", $ip_address)), "\n",
                   "    status=", $self->status_to_text($status), "\n",
                   "    unknown1=", udword_to_hex($unknown1), "\n",
                   "    unknown2=$unknown2\n",
                   "    unknown3=", udword_to_hex($unknown3), "\n",
                   "    unknown4=", udword_to_hex($unknown4), "\n");
#                  "    unknown5=", udword_to_hex($unknown5), "\n");
#                  "    unknown6=", udword_to_hex($unknown6), "\n",
#                  "    unknown7=", udword_to_hex($unknown7), "\n",
#                  "    unknown8=", udword_to_hex($unknown8), "\n");
    
    if (!$self->verify_password($user_uin, $password))
    {
      $packet = "";

      $self->{'user_uin'} = $user_uin;
      
      $self->send_packet($self->construct_packet("BAD_LOGIN", $packet));
      return;
    }
    
    if (!defined($self->db_user_info($user_uin)))
    {
      # User not in database
      $self->db_create_new_user($user_uin);
    }
  
    if (!$self->session_check_user_status($user_uin, 'OFFLINE'))
    {
      print "User already online\n";

      $packet = "";
      $packet = insert_uword ($packet, 0x006E);
      $packet = insert_udword($packet, $user_uin);
      $packet = insert_udword($packet, 0);
      
      $self->send_packet($self->construct_packet("TRY_AGAIN", $packet));
      # Just a guess, but the 0x006E above looks very much like the code for
      # the server command USER_ONLINE...
    }
    else
    {
      $self->{'user_uin'} = $user_uin;
      
      $self->{'sessions'}{$user_uin}{'sequence_number'}  = 1;
      $self->{'sessions'}{$user_uin}{'sequence_number2'} = 1;
      $self->{'sessions'}{$user_uin}{'ip_address'} = 
        $self->{'socket'}->peerhost;
      $self->{'sessions'}{$user_uin}{'udp_port'} = $self->{'socket'}->peerport;
      $self->{'sessions'}{$user_uin}{'real_ip_address'} = $ip_address;
      $self->{'sessions'}{$user_uin}{'tcp_port'} = $tcp_port;
      $self->{'sessions'}{$user_uin}{'session_id'} = 
        $self->{'packet_header'}{'session_id'};
      $self->{'sessions'}{$user_uin}{'last_packet_received_time'} = time;

      $packet = "";
      $packet = insert_udword($packet, 0x0000008C);
      $packet = insert_uword ($packet, 0x00F0);
      $packet = insert_uword ($packet, 0x000A);
      $packet = insert_uword ($packet, 0x000A);
      $packet = insert_uword ($packet, 0x0005);
      $packet = insert_ipaddr($packet, $ip_address);
      if (1)
      {
        $packet = insert_udword($packet, 0x12ACD0AC);
      }
      else
      {
        $packet = insert_udword($packet, 0x12ACf4AC);
        $packet = insert_udword($packet, 0x0c000296);
        $packet = insert_udword($packet, 0x27702037);
        $packet = insert_udword($packet, 0xaa3abe13);
        $packet = insert_uword ($packet, 0x0037);
      }

      $self->send_packet($self->construct_packet("LOGIN_REPLY", $packet, 0,1));

      $self->session_set_user_status($user_uin, $status);
    }
  }
}

sub receive_new_user_1
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($unknown);
  ($packet, $unknown) = extract_udword($packet) if ($version == 5);


  $DEBUG && warn("<<NEW_USER_1:\n",
                 ($version == 5 ? "    unknown=" . udword_to_hex($unknown) :
                  ""));
}

sub receive_old_user_reg
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($reg_seq_num, $user_uin, $password, $unknown);
  $reg_seq_num = 0;
  ($packet, $reg_seq_num) = extract_uword($packet) if ($version == 2);
  ($packet, $user_uin)    = extract_udword($packet);
  ($packet, $password)    = extract_string($packet);
  ($packet, $unknown)     = extract_udword($packet);


  $DEBUG && warn("<<OLD_USER_REG:\n",
                 "    reg_seq_num=$reg_seq_num\n",
                 "    user_uin=$user_uin\n",
                 "    password=$password\n",
                 "    unknown=", udword_to_hex($unknown), "\n");

  if (!$self->verify_password($user_uin, $password))
  {
    $packet = "";
    $packet = insert_uword ($packet, $reg_seq_num) if ($version == 2);
    $packet = insert_udword($packet, $user_uin);
    
    $self->send_packet($self->construct_packet("BAD_LOGIN", $packet));
    return;
  }

  if (!defined($self->db_user_info($user_uin)))
  {
    # User not in database
    $self->db_create_new_user($user_uin);
  }

  $self->{'user_uin'} = 0;

  if ($version == 2)
  {
    $packet = "";
    $packet = insert_uword ($packet, $reg_seq_num) if ($version == 2);
    $packet = insert_udword($packet, $user_uin);
    
    $self->send_packet($self->construct_packet("OLD_USER_UIN", $packet));
  }
  elsif ($version == 5)
  {
    my $packet_header = $self->{'packet_header'};
    
    # Change to a version 3 reply
    $self->{'packet_header'}->{'version'} = 3;
    $self->{'user_uin'} = $user_uin;


    $packet = "";
    
    $self->send_packet($self->construct_packet("OLD_USER_UIN", $packet, 0, 
                                               $packet_header->{'seq_num2'}));
  }
}

sub receive_new_user_reg
{
  my $self   = shift;
  my $packet = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($reg_seq_num, $password, $unknown1, $unknown2);
  $reg_seq_num = 0;
  ($packet, $reg_seq_num) = extract_uword($packet) if ($version == 2);
  ($packet, $password)    = extract_string($packet);
  ($packet, $unknown1)    = extract_udword($packet);
  ($packet, $unknown2)    = extract_udword($packet);


  $DEBUG && warn("<<NEW_USER_REG:\n",
                 "    reg_seq_num=$reg_seq_num\n",
                 "    password=$password\n",
                 "    unknown1=$unknown1\n",
                 "    unknown2=$unknown2\n");

  # We use the password field to pass the user's login name
  my($name, $passwd, $uid, $gid, $quota, $comment, $gcos, $dir, $shell) =
    getpwnam($password);
  my($user_uin) = $uid + $config::base_uin;

  
  if (!defined($name))
  {
    # No such user
    $packet = "";
    $packet = insert_uword ($packet, $reg_seq_num) if ($version == 2);
    $packet = insert_udword($packet, $user_uin);
    
    $self->send_packet($self->construct_packet("BAD_LOGIN", $packet));
    return;
  }
  
  if (!defined($self->db_user_info($user_uin)))
  {
    # User not in database
    $self->db_create_new_user($user_uin);
  }
  
  
  if ($version == 2)
  {
    $self->send_packet($self->construct_packet("NEW_USER_UIN",
                                               pack("vVV",
                                                    $reg_seq_num,
                                                    $user_uin,
                                                    $unknown2)));
  }
  elsif ($version == 5)
  {
    my $packet_header = $self->{'packet_header'};
    
    # Change to a version 3 reply
    $self->{'packet_header'}->{'version'} = 3;
    $self->{'user_uin'} = $user_uin;
    
    
    $packet = "";
    
    $self->send_packet($self->construct_packet("NEW_USER_UIN", $packet, 0, 
                                               $packet_header->{'seq_num2'}));
  }
}

sub receive_new_user_info
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};
  
  my($user_uin, $info_seq_num, $nick_name, $first_name, $last_name, $email,
     $authorization_status);
  $info_seq_num = 0;
  ($packet, $user_uin)             = $self->extract_or_get_user_uin($packet);
  ($packet, $info_seq_num)         = extract_uword($packet) if ($version == 2);
  ($packet, $nick_name)            = extract_string($packet);
  ($packet, $first_name)           = extract_string($packet);
  ($packet, $last_name)            = extract_string($packet);
  ($packet, $email)                = extract_string($packet);
  ($packet, $authorization_status) = extract_byte($packet);


  $DEBUG && warn("<<NEW_USER_INFO:\n",
                 "    user_uin=$user_uin\n",
                 "    info_seq_num=$info_seq_num\n",
                 "    nick_name=$nick_name\n",
                 "    first_name=$first_name\n",
                 "    last_name=$last_name\n",
                 "    email=$email\n",
                 "    authorization_status=$authorization_status\n");

  $self->{'user_uin'} = $user_uin;

  my($user_info) = $self->db_user_info($user_uin);
  
  if (defined($user_info))
  {
    ($$user_info{'nick_name'},
     $$user_info{'first_name'},
     $$user_info{'last_name'},
     $$user_info{'email'},
     $$user_info{'authorization'}) =
       ($nick_name, $first_name, $last_name, $email, $authorization_status);

    $self->db_update_user_info($user_uin, $user_info);

    $packet = "";
    $packet = insert_uword($packet, $info_seq_num) if ($version == 2);
    $self->send_packet($self->construct_packet("NEW_USER_REPLY", $packet));
  }
#  else
#  {
#    $self->send_packet($self->construct_packet("UPDATE_FAILED",
#                                               pack("v", $info_seq_num)));
#  }
}

sub receive_login_2
{
  my $self   = shift;
  my $packet = shift;

  my($user_uin, $unknown);
  ($packet, $user_uin) = $self->extract_or_get_user_uin($packet);
  ($packet, $unknown)  = extract_byte($packet);
  
  
  $DEBUG && warn("<<LOGIN_2:\n",
                 "    user_uin=$user_uin\n",
                 "    unknown=$unknown\n");

  $self->{'user_uin'} = $user_uin;
}

sub receive_info_req
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin, $info_seq_num, $searched_uin);
  $info_seq_num = 0;
  ($packet, $user_uin)     = $self->extract_or_get_user_uin($packet);
  ($packet, $info_seq_num) = extract_uword($packet) if ($version == 2);
  ($packet, $searched_uin) = extract_udword($packet);
  
  
  $DEBUG && warn("<<INFO_REQ:\n",
                 "    user_uin=$user_uin\n",
                 "    info_seq_num=$info_seq_num\n",
                 "    searched_uin=$searched_uin\n");

  $self->{'user_uin'} = $user_uin;

  my($user_info) = $self->db_user_info($searched_uin);
  
  if (defined($user_info))
  {
    my($nick_name, $first_name, $last_name, $email, $authorization_status) =
      ($$user_info{'nick_name'},
       $$user_info{'first_name'},
       $$user_info{'last_name'},
       $$user_info{'email'},
       $$user_info{'authorization'});
    
    $packet = "";
    $packet = insert_uword ($packet, $info_seq_num) if ($version == 2);
    $packet = insert_udword($packet, $searched_uin);
    $packet = insert_string($packet, $nick_name);
    $packet = insert_string($packet, $first_name);
    $packet = insert_string($packet, $last_name);
    $packet = insert_string($packet, $email);
    $packet = insert_byte  ($packet, $authorization_status);
    
    $self->send_packet($self->construct_packet("INFO_REPLY", $packet));
  }
  else
  {
    print "INFO_REQ: Don't know what to send\n";
  }
}

sub receive_ext_info_req
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin, $info_seq_num, $searched_uin);
  $info_seq_num = 0;
  ($packet, $user_uin)     = $self->extract_or_get_user_uin($packet);
  ($packet, $info_seq_num) = extract_uword($packet) if ($version == 2);
  ($packet, $searched_uin) = extract_udword($packet);
  

  $DEBUG && warn("<<EXT_INFO_REQ:\n",
                 "    user_uin=$user_uin\n",
                 "    info_seq_num=$info_seq_num\n",
                 "    searched_uin=$searched_uin\n");

  $self->{'user_uin'} = $user_uin;

  my($user_info) = $self->db_user_info($searched_uin);

  if (defined($user_info))
  {
    my($city, $country_code, $timezone, $state, $age, $sex, $phone, 
       $homepage, $about) =
         ($$user_info{'home_city'},
          $$user_info{'home_country_code'},
          $$user_info{'timezone'},
          $$user_info{'home_state'},
          $$user_info{'age'},
          $$user_info{'sex'},
          $$user_info{'home_phone'},
          $$user_info{'homepage'},
          $$user_info{'about'});
    
    $packet = "";
    $packet = insert_uword ($packet, $info_seq_num) if ($version == 2);
    $packet = insert_udword($packet, $searched_uin);
    $packet = insert_string($packet, $city);
    $packet = insert_uword ($packet, $country_code);
    $packet = insert_byte  ($packet, $timezone);
    $packet = insert_string($packet, $state);
    $packet = insert_uword ($packet, $age);
    $packet = insert_byte  ($packet, $sex);
    $packet = insert_string($packet, $phone);
    $packet = insert_string($packet, $homepage);
    $packet = insert_string($packet, $about);
    
    $self->send_packet($self->construct_packet("EXT_INFO_REPLY", $packet));
  }
  else
  {
    print "EXT_INFO_REQ: Don't know what to send\n";
  }
}

sub receive_update_info
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin, $info_seq_num, $nick_name, $first_name, $last_name, $email);
  $info_seq_num = 0;
  ($packet, $user_uin)     = $self->extract_or_get_user_uin($packet);
  ($packet, $info_seq_num) = extract_uword($packet) if ($version == 2);
  ($packet, $nick_name)    = extract_string($packet);
  ($packet, $first_name)   = extract_string($packet);
  ($packet, $last_name)    = extract_string($packet);
  ($packet, $email)        = extract_string($packet);
  

  $DEBUG && warn("<<UPDATE_INFO:\n",
                 "    user_uin=$user_uin\n",
                 "    info_seq_num=$info_seq_num\n",
                 "    nick_name=$nick_name\n",
                 "    first_name=$first_name\n",
                 "    last_name=$last_name\n",
                 "    email=$email\n");

  $self->{'user_uin'} = $user_uin;

  my($user_info) = $self->db_user_info($user_uin);
  
  if (defined($user_info))
  {
    ($$user_info{'nick_name'},
     $$user_info{'first_name'},
     $$user_info{'last_name'},
     $$user_info{'email'}) =
       ($nick_name, $first_name, $last_name, $email);

    $self->db_update_user_info($user_uin, $user_info);

    $packet = "";
    $packet = insert_uword($packet, $info_seq_num) if ($version == 2);
    $self->send_packet($self->construct_packet("UPDATE_SUCCEEDED", $packet));
  }
  else
  {
    $packet = "";
    $packet = insert_uword($packet, $info_seq_num) if ($version == 2);
    $self->send_packet($self->construct_packet("UPDATE_FAILED", $packet));
  }
}

sub receive_update_ext_info
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin, $info_seq_num, $city, $country_code, $timezone, $state, 
     $age, $sex, $phone, $homepage, $about);
  $info_seq_num = 0;
  ($packet, $user_uin)       = $self->extract_or_get_user_uin($packet);
  ($packet, $info_seq_num)   = extract_uword ($packet) if ($version == 2);
  ($packet, $city)           = extract_string($packet);
  ($packet, $country_code)   = extract_uword ($packet);
  ($packet, $timezone)       = extract_byte  ($packet);
  ($packet, $state)          = extract_string($packet);
  ($packet, $age)            = extract_uword ($packet);
  ($packet, $sex)            = extract_byte  ($packet);
  ($packet, $phone)          = extract_string($packet);
  ($packet, $homepage)       = extract_string($packet);
  ($packet, $about)          = extract_string($packet);


  $DEBUG && warn("<<UPDATE_EXT_INFO:\n",
                 "    user_uin=$user_uin\n",
                 "    info_seq_num=$info_seq_num\n",
                 "    city=$city\n",
                 "    country_code=$country_code\n",
                 "    timezone=$timezone\n",
                 "    state=$state\n",
                 "    age=$age\n",
                 "    sex=$sex\n",
                 "    phone=$phone\n",
                 "    homepage=$homepage\n",
                 "    about=$about\n");
                 
  $self->{'user_uin'} = $user_uin;

  my($user_info) = $self->db_user_info($user_uin);
  
  if (defined($user_info))
  {
    ($$user_info{'city'},
     $$user_info{'country_code'},
     $$user_info{'timezone'},
     $$user_info{'state'},
     $$user_info{'age'},
     $$user_info{'sex'},
     $$user_info{'phone'},
     $$user_info{'homepage'},
     $$user_info{'about'}) =
       ($city, $country_code, $timezone, $state, $age, $sex, $phone, 
        $homepage, $about);
    
    $self->db_update_user_info($user_uin, $user_info);

    $packet = "";
    $packet = insert_uword($packet, $info_seq_num) if ($version == 2);
    $self->send_packet($self->construct_packet("UPDATE_EXT_REPLY", $packet));
  }
  else
  {
    print "UPDATE_EXT_INFO: Don't know what to send\n";
  }
}

sub receive_auth_update
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin, $info_seq_num, $authorization_status);
  $info_seq_num = 0;
  ($packet, $user_uin)             = $self->extract_or_get_user_uin($packet);
  ($packet, $info_seq_num)         = extract_uword($packet) if ($version == 2);
  ($packet, $authorization_status) = extract_byte($packet);
  
  
  $DEBUG && warn("<<AUTH_UPDATE:\n",
                 "    user_uin=$user_uin\n",
                 "    info_seq_num=$info_seq_num\n",
                 "    authorization_status=$authorization_status\n");

  $self->{'user_uin'} = $user_uin;

  my($user_info) = $self->db_user_info($user_uin);
  
  if (defined($user_info))
  {
    $$user_info{'authorization'} = $authorization_status;
    
    $self->db_update_user_info($user_uin, $user_info);

    $packet = "";
    $packet = insert_uword($packet, $info_seq_num) if ($version == 2);
    $self->send_packet($self->construct_packet("AUTH_REPLY", $packet));
  }
  else
  {
    print "AUTH_UPDATE: Don't know what to send\n";
  }
}

sub receive_request_offline_messages
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my $user_uin;
  ($packet, $user_uin) = $self->extract_or_get_user_uin($packet);
  
  
  $DEBUG && warn("<<REQUEST_OFFLINE_MESSAGES:\n",
                 "    user_uin=$user_uin\n");

  $self->{'user_uin'} = $user_uin;

  my($user_info) = $self->db_user_info($user_uin);
  
  if (defined($user_info))
  {
    my $sth = $self->db_do("SELECT * FROM messages WHERE uin = $user_uin;");
    while (my $pointer = $sth->fetchrow_hashref)
    {
      my $sender_uin   = $pointer->{'sender_uin'};
      my $message_type = $pointer->{'message_type'};
      my $message_text = $pointer->{'message'};
      
      # TIMESTAMP format: YYYYMMDDHHMMSS      
      my $timestamp = $pointer->{'timestamp'};
      my($year, $month, $day, $hour, $min, $sec) =
        ($timestamp =~ /(\d{4})(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)/);
      print "$year, $month, $day, $hour, $min, $sec\n";

      # FIXME: Has to do withthe timezone... This is wrong :)
      $hour-- if ($hour > 0);
      
      $packet = "";
      $packet = insert_udword($packet, $sender_uin);
      $packet = insert_uword ($packet, $year);
      $packet = insert_byte  ($packet, $month);
      $packet = insert_byte  ($packet, $day);
      $packet = insert_byte  ($packet, $hour);
      $packet = insert_byte  ($packet, $min);
      $packet = insert_uword ($packet, $message_type);
      $packet = insert_string($packet, $message_text);
      
      $self->send_packet($self->construct_packet("RECEIVE_MESSAGE", $packet));
    }
    $sth->finish;    
  }
  
  $self->send_packet($self->construct_packet("END_OFFLINE_MESSAGES"));
}

sub receive_ack_offline_messages
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin);
  ($packet, $user_uin) = $self->extract_or_get_user_uin($packet);
  

  $DEBUG && warn("<<ACK_OFFLINE_MESSAGES:\n",
                 "    user_uin=$user_uin\n");

  $self->{'user_uin'} = $user_uin;

  my($user_info) = $self->db_user_info($user_uin);
  
  if (defined($user_info))
  {
    my $sth = $self->db_do("DELETE FROM messages WHERE uin = $user_uin;");
  }
}

sub receive_query_addons
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin);
  ($packet, $user_uin) = $self->extract_or_get_user_uin($packet);

  
  $DEBUG && warn("<<QUERY_ADDONS:\n",
                 "    user_uin=$user_uin\n");

  $self->{'user_uin'} = $user_uin;
}

sub receive_query_servers
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin);
  ($packet, $user_uin) = $self->extract_or_get_user_uin($packet);

  
  $DEBUG && warn("<<QUERY_SERVERS:\n",
                 "    user_uin=$user_uin\n");

  $self->{'user_uin'} = $user_uin;
}

sub receive_keep_alive
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin);
  ($packet, $user_uin) = $self->extract_or_get_user_uin($packet);

  
  $DEBUG && $DEBUG_ACK && warn("<<KEEP_ALIVE:\n",
                               "    user_uin=$user_uin\n");

  $self->{'user_uin'} = $user_uin;
}

sub receive_keep_alive_2
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin);
  ($packet, $user_uin) = $self->extract_or_get_user_uin($packet);


  $DEBUG && $DEBUG_ACK && warn("<<KEEP_ALIVE_2:\n",
                               "    user_uin=$user_uin\n");

  $self->{'user_uin'} = $user_uin;
}

sub receive_contact_list
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin, $num_contacts);
  ($packet, $user_uin)     = $self->extract_or_get_user_uin($packet);
  ($packet, $num_contacts) = extract_byte($packet);


  $DEBUG && warn("<<CONTACT_LIST:\n",
                 "    user_uin=$user_uin\n",
                 "    num_contacts=$num_contacts\n");

  $self->{'user_uin'} = $user_uin;

  while ($num_contacts--)
  {
    my($contact_uin);
    ($packet, $contact_uin) = extract_udword($packet);

    $DEBUG && warn("    contact_uin=$contact_uin\n");

    $self->{'sessions'}{$contact_uin}{'notified'}{$user_uin} = 1;

    $self->notify_user($contact_uin, $user_uin);
  }


  # Tramsmit offline messages
  my($user_info) = $self->db_user_info($user_uin);  
  if (defined($user_info))
  {
    my $sth = $self->db_do("SELECT * FROM messages WHERE uin = $user_uin;");
    while (my $pointer = $sth->fetchrow_hashref)
    {
      my $sender_uin   = $pointer->{'sender_uin'};
      my $message_type = $pointer->{'message_type'};
      my $message_text = $pointer->{'message'};
      
      # TIMESTAMP format: YYYYMMDDHHMMSS      
      my $timestamp = $pointer->{'timestamp'};
      my($year, $month, $day, $hour, $min, $sec) =
        ($timestamp =~ /(\d{4})(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)/);
      print "$year, $month, $day, $hour, $min, $sec\n";

      # FIXME: Has to do with the timezone... This is wrong :)
      $hour-- if ($hour > 0);
      
      $packet = "";
      $packet = insert_udword($packet, $sender_uin);
      $packet = insert_uword ($packet, $year);
      $packet = insert_byte  ($packet, $month);
      $packet = insert_byte  ($packet, $day);
      $packet = insert_byte  ($packet, $hour);
      $packet = insert_byte  ($packet, $min);
      $packet = insert_uword ($packet, $message_type);
      $packet = insert_string($packet, $message_text);
      
      $self->send_packet($self->construct_packet("RECEIVE_MESSAGE", $packet));
    }
    $sth->finish;    
  }

  my $packet1 =
    $self->construct_packet("END_CONTACTLIST_STATUS",
                            "",
                            $self->{'sessions'}{$user_uin}{'sequence_number'}++,
                            0);
  my $packet2 =
    $self->construct_packet("END_OFFLINE_MESSAGES",
                            "",
                            $self->{'sessions'}{$user_uin}{'sequence_number'}++,
                            0);

  $self->send_multi_packet($packet1, $packet2);
}

sub receive_status_change
{
  my $self = shift;
  my $packet = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin, $status);
  ($packet, $user_uin) = $self->extract_or_get_user_uin($packet);
  ($packet, $status) = extract_udword($packet);


  $DEBUG && warn("<<STATUS_CHANGE:\n",
                 "    user_uin=$user_uin\n",
                 "    status=", $self->status_to_text($status), "\n");

  $self->{'user_uin'} = $user_uin;
  $self->session_set_user_status($user_uin, $status);
  
  # We have to increase sequence_number2 for some strange reason
  $self->{'sessions'}{$user_uin}{'sequence_number2'}++;
}

sub receive_search_uin
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin, $search_seq_num, $searched_uin);
  ($packet, $user_uin)       = $self->extract_or_get_user_uin($packet);
  ($packet, $search_seq_num) = extract_uword($packet);
  ($packet, $searched_uin)   = extract_udword($packet);


  $DEBUG && warn("<<SEARCH_UIN:\n",
                 "    user_uin=$user_uin\n",
                 "    search_seq_num=$search_seq_num\n",
                 "    searched_uin=$searched_uin\n");

  $self->{'user_uin'} = $user_uin;
  
  my($more_found) = 0;  
  
  my($user_info) = $self->db_user_info($searched_uin);

  if (defined($user_info))
  {
    my($nick_name, $first_name, $last_name, $email, $authorization_status) =
      ($$user_info{'nick_name'},
       $$user_info{'first_name'},
       $$user_info{'last_name'},
       $$user_info{'email'},
       $$user_info{'authorization'});

    $packet = "";
    $packet = insert_uword ($packet, $search_seq_num);
    $packet = insert_udword($packet, $searched_uin);
    $packet = insert_string($packet, $nick_name);
    $packet = insert_string($packet, $first_name);
    $packet = insert_string($packet, $last_name);
    $packet = insert_string($packet, $email);
    $packet = insert_byte  ($packet, $authorization_status);
    
    $self->send_packet($self->construct_packet("USER_FOUND", $packet));
  }
  
  $packet = "";
  $packet = insert_uword ($packet, $search_seq_num);
  $packet = insert_byte  ($packet, $more_found);
  $self->send_packet($self->construct_packet("END_OF_SEARCH", $packet));
}

sub receive_search_user
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin, $search_seq_num, $nick_name, $first_name, $last_name, $email);
  $search_seq_num = 0;
  ($packet, $user_uin)       = $self->extract_or_get_user_uin($packet);
  ($packet, $search_seq_num) = extract_uword ($packet) if ($version == 2);
  ($packet, $nick_name)      = extract_string($packet);
  ($packet, $first_name)     = extract_string($packet);
  ($packet, $last_name)      = extract_string($packet);
  ($packet, $email)          = extract_string($packet);


  $DEBUG && warn("<<SEARCH_USER:\n",
                 "    user_uin=$user_uin\n",
                 "    search_seq_num=$search_seq_num\n",
                 "    nick_name=$nick_name\n",
                 "    first_name=$first_name\n",
                 "    last_name=$last_name\n",
                 "    email=$email\n");

  $self->{'user_uin'} = $user_uin;

  my($more_found, @keys);
  if ($email)
  {
    ($more_found, @keys) = $self->db_exact_search_by_email($email);
  }
  else
  {
    ($more_found, @keys) =
      $self->db_exact_search_by_name($nick_name, $first_name, $last_name);
  }

  foreach my $searched_uin (@keys)
  {
    my($user_info) = $self->db_user_info($searched_uin);

    if (defined($user_info))
    {
      my($nick_name, $first_name, $last_name, $email, $authorization_status) =
        ($$user_info{'nick_name'},
         $$user_info{'first_name'},
         $$user_info{'last_name'},
         $$user_info{'email'},
         $$user_info{'authorization'});

      $packet = "";
      $packet = insert_uword ($packet, $search_seq_num);
      $packet = insert_udword($packet, $searched_uin);
      $packet = insert_string($packet, $nick_name);
      $packet = insert_string($packet, $first_name);
      $packet = insert_string($packet, $last_name);
      $packet = insert_string($packet, $email);
      $packet = insert_byte  ($packet, $authorization_status);
      
      $self->send_packet($self->construct_packet("USER_FOUND", $packet));
    }
  }

  $packet = "";
  $packet = insert_uword ($packet, $search_seq_num);
  $packet = insert_byte  ($packet, $more_found);
  $self->send_packet($self->construct_packet("END_OF_SEARCH", $packet));
}

sub receive_add_to_list
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin, $added_uin);
  ($packet, $user_uin)  = $self->extract_or_get_user_uin($packet);
  ($packet, $added_uin) = extract_udword($packet);


  $DEBUG && warn("<<ADD_TO_LIST:\n",
                 "    user_uin=$user_uin\n",
                 "    added_uin=$added_uin\n");

  $self->{'user_uin'} = $user_uin;

  $self->{'sessions'}{$added_uin}{'notified'}{$user_uin} = 1;
  $self->notify_user($added_uin, $user_uin);
}

sub receive_send_message
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin, $receiver_uin, $message_type, $message);
  ($packet, $user_uin)     = $self->extract_or_get_user_uin($packet);
  ($packet, $receiver_uin) = extract_udword($packet);
  ($packet, $message_type) = extract_uword($packet);
  ($packet, $message)      = extract_string($packet);
  

  $DEBUG && warn("<<SEND_MESSAGE:\n",
                 "    user_uin=$user_uin\n",
                 "    receiver_uin=$receiver_uin\n",
                 "    message_type=$message_type\n",
                 "    message=$message\n");

  $self->message_to_user($user_uin, $receiver_uin, $message_type, $message);
}

sub receive_send_text_code
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin, $text_code, $unknown);
  ($packet, $user_uin)  = $self->extract_or_get_user_uin($packet);
  ($packet, $text_code) = extract_string($packet);
  ($packet, $unknown)   = extract_uword($packet); 


  $DEBUG && warn("<<SEND TEXT CODE:\n",
                 "    user_uin=$user_uin\n",
                 "    text_code=$text_code\n",
                 "    unknown=", uword_to_hex($unknown), "\n");

  $self->{'user_uin'} = $user_uin;

  if ($text_code eq 'B_USER_DISCONNECTED')
  {
    if (!$self->session_check_user_status($user_uin, 'OFFLINE'))
    {
      if ($self->{'sessions'}{$user_uin}{'ip_address'} eq $self->{'socket'}->peerhost &&
          $self->{'sessions'}{$user_uin}{'udp_port'} == $self->{'socket'}->peerport)
      {
        $self->session_set_user_status($user_uin, $status_bits{'OFFLINE'});
        print "Disconnected: $user_uin\n";
      }
    }
  }
  else
  {
#    $self->send_ack($self->{'packet_header'}{'seq_num'});
  }
}

sub receive_req_add_to_list
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};
  
  my($user_uin, $receiver_uin, $message_type, $message);
  ($packet, $user_uin)     = $self->extract_or_get_user_uin($packet);
  ($packet, $receiver_uin) = extract_udword($packet);
  ($packet, $message_type) = extract_uword($packet);
  ($packet, $message)      = extract_string($packet);


  $DEBUG && warn("<<REQ_ADD_TO_LIST:\n",
                 "    user_uin=$user_uin\n",
                 "    receiver_uin=$receiver_uin\n",
                 "    message_type=$message_type\n",
                 "    message=$message\n");

  $self->message_to_user($user_uin, $receiver_uin, $message_type, $message);
}

sub receive_change_password
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin, $change_seq_num, $password); 
  ($packet, $user_uin)       = $self->extract_or_get_user_uin($packet);
  ($packet, $change_seq_num) = extract_uword($packet);
  ($packet, $password)       = extract_string($packet);


  $DEBUG && warn("<<CHANGE_PASSWORD:\n",
                 "    user_uin=$user_uin\n",
                 "    change_seq_num=$change_seq_num\n",
                 "    password=$password\n");

  $self->{'user_uin'} = $user_uin;
}

sub receive_delete_uin
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin, $delete_seq_num, $delete_uin, $password); 
  ($packet, $user_uin)       = $self->extract_or_get_user_uin($packet);
  ($packet, $delete_seq_num) = extract_uword($packet);
  ($packet, $delete_uin)     = extract_udword($packet);
  ($packet, $password)       = extract_string($packet);


  $DEBUG && warn("<<DELETE_UIN:\n",
                 "    user_uin=$user_uin\n",
                 "    delete_seq_num=$delete_seq_num\n",
                 "    delete_uin=$delete_uin\n",
                 "    password=$password\n");

  $self->{'user_uin'} = $user_uin;

  # FIXME: Add password check and return DELETE_FAILED if incorrect.

  if (defined($self->db_user_info($delete_uin)))
  {
    $packet = "";
    $packet = insert_uword ($packet, $delete_seq_num) if ($version == 2);
    $packet = insert_udword($packet, $delete_uin);
    $self->send_packet($self->construct_packet("DELETE_SUCCEEDED", $packet));
    
    # FIXME: Delete the user.
  }
  else
  {
    $packet = "";
    $packet = insert_uword ($packet, $delete_seq_num) if ($version == 2);
    $packet = insert_udword($packet, $delete_uin);
    $self->send_packet($self->construct_packet("DELETE_FAILED", $packet));
  }
}

sub receive_meta_user
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin, $meta_cmd);
  ($packet, $user_uin) = $self->extract_or_get_user_uin($packet);
  ($packet, $meta_cmd) = extract_uword($packet);

  $DEBUG && warn("<<META_USER:\n",
                 "    user_uin=$user_uin\n",
                 "    meta_cmd=", $client_meta_commands_bynumber{$meta_cmd},
                 " (", uword_to_hex($meta_cmd), ")\n");

  $self->{'user_uin'} = $user_uin;

  my $command_name = "receive_" . 
    lc($client_meta_commands_bynumber{$meta_cmd});
  my $coderef = $self->can($command_name);
  if ($command_name && $coderef)
  {
    # we've found a method that can process this type of packet
    return $self->$command_name($packet);
  }
  else
  {
    # TODO: can't cope - what do we do with this packet!!
    warn("UNKNOWN META COMMAND: '$meta_cmd/", &uword_to_hex($meta_cmd));
    return 0;
  }
}

sub receive_random_set
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};
  
  my($user_uin, $random_group);
  ($packet, $user_uin)     = $self->extract_or_get_user_uin($packet);
  ($packet, $random_group) = extract_udword($packet);


  $DEBUG && warn("<<RANDOM_SET:\n",
                 "    user_uin=$user_uin\n",
                 "    random_group=$random_group\n");

  # FIXME: Implement this
}

sub receive_random_search
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};
  
  my($user_uin, $random_group);
  ($packet, $user_uin)     = $self->extract_or_get_user_uin($packet);
  ($packet, $random_group) = extract_uword($packet);


  $DEBUG && warn("<<RANDOM_SEARCH:\n",
                 "    user_uin=$user_uin\n",
                 "    random_group=$random_group\n");

  # FIXME: Implement this
}

sub receive_invisible_list
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};
  
  my($user_uin, $num_users);
  ($packet, $user_uin)  = $self->extract_or_get_user_uin($packet);
  ($packet, $num_users) = extract_byte($packet);


  $DEBUG && warn("<<INVISIBLE_LIST:\n",
                 "    user_uin=$user_uin\n",
                 "    num_users=$num_users\n");

  $self->{'user_uin'} = $user_uin;

  while ($num_users--)
  {
    my($invis_uin);
    ($packet, $invis_uin) = extract_udword($packet);

    $DEBUG && warn("    invis_uin=$invis_uin\n");

    # FIXME: Implement this
  }
}

sub receive_visible_list
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};
  
  my($user_uin, $num_users);
  ($packet, $user_uin)  = $self->extract_or_get_user_uin($packet);
  ($packet, $num_users) = extract_byte($packet);


  $DEBUG && warn("<<VISIBLE_LIST:\n",
                 "    user_uin=$user_uin\n",
                 "    num_users=$num_users\n");

  $self->{'user_uin'} = $user_uin;

  while ($num_users--)
  {
    my($vis_uin);
    ($packet, $vis_uin) = extract_udword($packet);

    $DEBUG && warn("    vis_uin=$vis_uin\n");

    # FIXME: Implement this
  }
}

sub receive_visibility_lists
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};
  
  my($user_uin, $list_uin, $list, $list_status);
  ($packet, $user_uin)    = $self->extract_or_get_user_uin($packet);
  ($packet, $list_uin)    = extract_udword($packet);
  ($packet, $list)        = extract_byte  ($packet); # 1 = invis, 2 = visible
  ($packet, $list_status) = extract_byte  ($packet);


  $DEBUG && warn("<<VISIBILITY_LISTS:\n",
                 "    user_uin=$user_uin\n",
                 "    list_uin=$list_uin\n",
                 "    list=$list\n",
                 "    list_status=$list_status\n");

  # FIXME: Implement this
}

sub receive_meta_info_set_main
{
  my $self     = shift;
  my $packet   = shift;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($nick_name, $first_name, $last_name, $email, $secondary_email, 
     $old_email, $home_city, $home_state, $home_phone, $home_fax, 
     $home_street, $home_cellular, $home_zip_code,
     $home_country_code, $timezone, $publish_email_address);
  ($packet, $nick_name)             = extract_string($packet);
  ($packet, $first_name)            = extract_string($packet);
  ($packet, $last_name)             = extract_string($packet);
  ($packet, $email)                 = extract_string($packet);
  ($packet, $secondary_email)       = extract_string($packet);
  ($packet, $old_email)             = extract_string($packet);
  ($packet, $home_city)             = extract_string($packet);
  ($packet, $home_state)            = extract_string($packet);
  ($packet, $home_phone)            = extract_string($packet);
  ($packet, $home_fax)              = extract_string($packet);
  ($packet, $home_street)           = extract_string($packet);
  ($packet, $home_cellular)         = extract_string($packet);
  ($packet, $home_zip_code)         = extract_udword($packet);
  ($packet, $home_country_code)     = extract_uword ($packet);
  ($packet, $timezone)              = extract_byte  ($packet);
  ($packet, $publish_email_address) = extract_byte  ($packet);


  $DEBUG && warn("    nick_name=$nick_name\n",
                 "    first_name=$first_name\n",
                 "    last_name=$last_name\n",
                 "    email=$email\n",
                 "    secondary_email=$secondary_email\n",
                 "    old_email=$old_email\n",
                 "    home_city=$home_city\n",
                 "    home_state=$home_state\n",
                 "    home_phone=$home_phone\n",
                 "    home_fax=$home_fax\n",
                 "    home_street=$home_street\n",
                 "    home_cellular=$home_cellular\n",
                 "    home_zip_code=$home_zip_code\n",
                 "    home_country_code=$home_country_code\n",
                 "    timezone=$timezone\n",
                 "    publish_email_address=$publish_email_address\n");

  my($user_info) = $self->db_user_info($user_uin);
  
  if (defined($user_info))
  {
    $$user_info{'nick_name'}             = $nick_name;
    $$user_info{'first_name'}            = $first_name;
    $$user_info{'last_name'}             = $last_name;
    $$user_info{'email'}                 = $email;
    $$user_info{'secondary_email'}       = $secondary_email;
    $$user_info{'old_email'}             = $old_email;
    $$user_info{'home_city'}             = $home_city;
    $$user_info{'home_state'}            = $home_state;
    $$user_info{'home_phone'}            = $home_phone;
    $$user_info{'home_fax'}              = $home_fax;
    $$user_info{'home_street'}           = $home_street;
    $$user_info{'home_cellular'}         = $home_cellular;
    $$user_info{'home_zip_code'}         = $home_zip_code;
    $$user_info{'home_country_code'}     = $home_country_code; 
    $$user_info{'timezone'}              = $timezone;
    $$user_info{'publish_email_address'} = $publish_email_address;
    
    $self->db_update_user_info($user_uin, $user_info);
    
    $packet = "";
    $packet = insert_uword($packet,
                           $server_meta_commands{'META_SRV_SET_MAIN_REPLY'});
    $packet = insert_byte ($packet, 0x0A);
    
    $self->send_packet($self->construct_packet("META_USER_REPLY", $packet));
  }
  else
  {
    print "INFO_REQ: Don't know what to send\n";
  }
}

sub receive_meta_info_set_more
{
  my $self     = shift;
  my $packet   = shift;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($age, $sex, $homepage, $birth_year, $birth_month, $birth_day,
     $language_1, $language_2, $language_3);
  ($packet, $age)            = extract_uword ($packet);
  ($packet, $sex)            = extract_byte  ($packet);
  ($packet, $homepage)       = extract_string($packet);
  ($packet, $birth_year)     = extract_byte  ($packet);
  ($packet, $birth_month)    = extract_byte  ($packet);
  ($packet, $birth_day)      = extract_byte  ($packet);
  ($packet, $language_1)     = extract_byte  ($packet);
  ($packet, $language_2)     = extract_byte  ($packet);
  ($packet, $language_3)     = extract_byte  ($packet);


  $DEBUG && warn("    age=$age\n",
                 "    sex=$sex\n",
                 "    homepage=$homepage\n",
                 "    birth_year=$birth_year\n",
                 "    birth_month=$birth_month\n",
                 "    birth_day=$birth_day\n",
                 "    language_1=$language_1\n",
                 "    language_2=$language_2\n",
                 "    language_3=$language_3\n");

  my($user_info) = $self->db_user_info($user_uin);
  
  if (defined($user_info))
  {
    $$user_info{'age'}         = $age;
    $$user_info{'sex'}         = $sex;
    $$user_info{'homepage'}    = $homepage;
    $$user_info{'birth_year'}  = $birth_year;
    $$user_info{'birth_month'} = $birth_month;
    $$user_info{'birth_day'}   = $birth_day;
    $$user_info{'language_1'}  = $language_1;
    $$user_info{'language_2'}  = $language_2;
    $$user_info{'language_3'}  = $language_3;
    
    $self->db_update_user_info($user_uin, $user_info);
    
    $packet = "";
    $packet = insert_uword($packet,
                           $server_meta_commands{'META_SRV_SET_MORE_REPLY'});
    $packet = insert_byte ($packet, 0x0A);
    
    $self->send_packet($self->construct_packet("META_USER_REPLY", $packet));
  }
  else
  {
    print "INFO_REQ: Don't know what to send\n";
  }
}

sub receive_meta_info_set_work
{
  my $self     = shift;
  my $packet   = shift;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($work_city, $work_state, $work_phone, $work_fax, $work_street, 
     $work_zip_code, $work_country_code, $work_company, $work_division, 
     $work_position, $work_occupation, $work_homepage);
  ($packet, $work_city)         = extract_string($packet);
  ($packet, $work_state)        = extract_string($packet);
  ($packet, $work_phone)        = extract_string($packet);
  ($packet, $work_fax)          = extract_string($packet);
  ($packet, $work_street)       = extract_string($packet);
  ($packet, $work_zip_code)     = extract_udword($packet);
  ($packet, $work_country_code) = extract_uword ($packet);
  ($packet, $work_company)      = extract_string($packet);
  ($packet, $work_division)     = extract_string($packet);
  ($packet, $work_position)     = extract_string($packet);
  ($packet, $work_occupation)   = extract_uword ($packet);
  ($packet, $work_homepage)     = extract_string($packet);


  $DEBUG && warn("    work_city=$work_city\n",
                 "    work_state=$work_state\n",
                 "    work_phone=$work_phone\n",
                 "    work_fax=$work_fax\n",
                 "    work_street=$work_street\n",
                 "    work_zip_code=$work_zip_code\n",
                 "    work_country_code=$work_country_code\n",
                 "    work_company=$work_company\n",
                 "    work_division=$work_division\n",
                 "    work_position=$work_position\n",
                 "    work_occupation=$work_occupation\n",
                 "    work_homepage=$work_homepage\n");

  my($user_info) = $self->db_user_info($user_uin);
  
  if (defined($user_info))
  {
    $$user_info{'work_city'}         = $work_city;
    $$user_info{'work_state'}        = $work_state;
    $$user_info{'work_phone'}        = $work_phone;
    $$user_info{'work_fax'}          = $work_fax;
    $$user_info{'work_street'}       = $work_street;
    $$user_info{'work_zip_code'}     = $work_zip_code;
    $$user_info{'work_country_code'} = $work_country_code;
    $$user_info{'work_company'}      = $work_company;
    $$user_info{'work_division'}     = $work_division;
    $$user_info{'work_position'}     = $work_position;
    $$user_info{'work_occupation'}   = $work_occupation;
    $$user_info{'work_homepage'}     = $work_homepage;
    
    $self->db_update_user_info($user_uin, $user_info);

    $packet = "";
    $packet = insert_uword($packet,
                           $server_meta_commands{'META_SRV_SET_WORK_REPLY'});
    $packet = insert_byte ($packet, 0x0A);
    
    $self->send_packet($self->construct_packet("META_USER_REPLY", $packet));
  }
  else
  {
    print "INFO_REQ: Don't know what to send\n";
  }
}

sub receive_meta_info_set_affiliations
{
  my $self     = shift;
  my $packet   = shift;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($organizations,
     $organization_type_1, $organization_text_1,
     $organization_type_2, $organization_text_2,
     $organization_type_3, $organization_text_3,
     $backgrounds,
     $background_type_1, $background_text_1,
     $background_type_2, $background_text_2,
     $background_type_3, $background_text_3);
  ($packet, $backgrounds)         = extract_byte  ($packet);
  ($packet, $background_type_1)   = extract_uword ($packet);
  ($packet, $background_text_1)   = extract_string($packet);
  ($packet, $background_type_2)   = extract_uword ($packet);
  ($packet, $background_text_2)   = extract_string($packet);
  ($packet, $background_type_3)   = extract_uword ($packet);
  ($packet, $background_text_3)   = extract_string($packet);
  ($packet, $organizations)       = extract_byte  ($packet);
  ($packet, $organization_type_1) = extract_uword ($packet);
  ($packet, $organization_text_1) = extract_string($packet);
  ($packet, $organization_type_2) = extract_uword ($packet);
  ($packet, $organization_text_2) = extract_string($packet);
  ($packet, $organization_type_3) = extract_uword ($packet);
  ($packet, $organization_text_3) = extract_string($packet);


  $DEBUG && warn("    organizations=$organizations\n",
                 "    organization_type_1=$organization_type_1\n",
                 "    organization_text_1=$organization_text_1\n",
                 "    organization_type_2=$organization_type_2\n",
                 "    organization_text_2=$organization_text_2\n",
                 "    organization_type_3=$organization_type_3\n",
                 "    organization_text_3=$organization_text_3\n",
                 "    backgrounds=$backgrounds\n",
                 "    background_type_1=$background_type_1\n",
                 "    background_text_1=$background_text_1\n",
                 "    background_type_2=$background_type_2\n",
                 "    background_text_2=$background_text_2\n",
                 "    background_type_3=$background_type_3\n",
                 "    background_text_3=$background_text_3\n");

  my($user_info) = $self->db_user_info($user_uin);
  
  if (defined($user_info))
  {
    $$user_info{'organization_type_1'} = $organization_type_1;
    $$user_info{'organization_text_1'} = $organization_text_1;  
    $$user_info{'organization_type_2'} = $organization_type_2;
    $$user_info{'organization_text_2'} = $organization_text_2; 
    $$user_info{'organization_type_3'} = $organization_type_3;
    $$user_info{'organization_text_3'} = $organization_text_3; 
    $$user_info{'background_type_1'}   = $background_type_1;
    $$user_info{'background_text_1'}   = $background_text_1;
    $$user_info{'background_type_2'}   = $background_type_2;
    $$user_info{'background_text_2'}   = $background_text_2;
    $$user_info{'background_type_3'}   = $background_type_3;
    $$user_info{'background_text_3'}   = $background_text_3;

    $self->db_update_user_info($user_uin, $user_info);

    $packet = "";
    $packet =
      insert_uword($packet,
                   $server_meta_commands{'META_SRV_SET_AFFILIATIONS_REPLY'});
    $packet = insert_byte ($packet, 0x0A);
    
    $self->send_packet($self->construct_packet("META_USER_REPLY", $packet));
  }
  else
  {
    print "INFO_REQ: Don't know what to send\n";
  }
}

sub receive_meta_info_set_interests
{
  my $self     = shift;
  my $packet   = shift;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($interests,
     $interest_type_1, $interest_text_1,
     $interest_type_2, $interest_text_2,
     $interest_type_3, $interest_text_3,
     $interest_type_4, $interest_text_4);
  ($packet, $interests)       = extract_byte  ($packet);
  ($packet, $interest_type_1) = extract_uword ($packet);
  ($packet, $interest_text_1) = extract_string($packet);
  ($packet, $interest_type_2) = extract_uword ($packet);
  ($packet, $interest_text_2) = extract_string($packet);
  ($packet, $interest_type_3) = extract_uword ($packet);
  ($packet, $interest_text_3) = extract_string($packet);
  ($packet, $interest_type_4) = extract_uword ($packet);
  ($packet, $interest_text_4) = extract_string($packet);


  $DEBUG && warn("    interests=$interests\n",
                 "    interest_type_1=$interest_type_1\n",
                 "    interest_text_1=$interest_text_1\n",
                 "    interest_type_2=$interest_type_2\n",
                 "    interest_text_2=$interest_text_2\n",
                 "    interest_type_3=$interest_type_3\n",
                 "    interest_text_3=$interest_text_3\n",
                 "    interest_type_4=$interest_type_4\n",
                 "    interest_text_4=$interest_text_4\n");

  my($user_info) = $self->db_user_info($user_uin);
  
  if (defined($user_info))
  {
    $$user_info{'interest_type_1'} = $interest_type_1;
    $$user_info{'interest_text_1'} = $interest_text_1; 
    $$user_info{'interest_type_2'} = $interest_type_2;
    $$user_info{'interest_text_2'} = $interest_text_2;  
    $$user_info{'interest_type_3'} = $interest_type_3;
    $$user_info{'interest_text_3'} = $interest_text_3; 
    $$user_info{'interest_type_4'} = $interest_type_4;
    $$user_info{'interest_text_4'} = $interest_text_4;
    
    $self->db_update_user_info($user_uin, $user_info);

    $packet = "";
    $packet =
      insert_uword($packet,
                   $server_meta_commands{'META_SRV_SET_INTERESTS_REPLY'});
    $packet = insert_byte ($packet, 0x0A);
    
    $self->send_packet($self->construct_packet("META_USER_REPLY", $packet));
  }
  else
  {
    print "INFO_REQ: Don't know what to send\n";
  }
}

sub receive_meta_info_set_homepage_category
{
  my $self     = shift;
  my $packet   = shift;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($category_1, $category_2, $category_3, $homepage_description);
  ($packet, $category_1)           = extract_byte  ($packet);
  ($packet, $category_2)           = extract_byte  ($packet);
  ($packet, $category_3)           = extract_byte  ($packet);
  ($packet, $homepage_description) = extract_string($packet);


  $DEBUG && warn("    category_1=$category_1\n",
                 "    category_2=$category_2\n",
                 "    category_3=$category_3\n",
                 "    category_text=$homepage_description\n");

  my($user_info) = $self->db_user_info($user_uin);
  
  if (defined($user_info))
  {
    $$user_info{'category_1'}    = $category_1;
    $$user_info{'category_2'}    = $category_2;
    $$user_info{'category_3'}    = $category_3;
    $$user_info{'category_text'} = $homepage_description;

    
    $self->db_update_user_info($user_uin, $user_info);

    $packet = "";
    $packet = insert_uword($packet,
                $server_meta_commands{'META_SRV_SET_HOMEPAGE_CATEGORY_REPLY'});
    $packet = insert_byte ($packet, 0x0A);

    $self->send_packet($self->construct_packet("META_USER_REPLY", $packet));
  }
  else
  {
    print "INFO_REQ: Don't know what to send\n";
  }
}

sub receive_meta_info_set_about
{
  my $self     = shift;
  my $packet   = shift;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($about);
  ($packet, $about) = extract_string($packet);


  $DEBUG && warn("    about=$about\n");

  my($user_info) = $self->db_user_info($user_uin);
  
  if (defined($user_info))
  {
    $$user_info{'about'} = $about;

    $self->db_update_user_info($user_uin, $user_info);

    $packet = "";
    $packet = insert_uword($packet,
                           $server_meta_commands{'META_SRV_SET_ABOUT_REPLY'});
    $packet = insert_byte ($packet, 0x0A);

    $self->send_packet($self->construct_packet("META_USER_REPLY", $packet));
  }
  else
  {
    print "INFO_REQ: Don't know what to send\n";
  }
}

sub receive_meta_info_set_security
{
  my $self     = shift;
  my $packet   = shift;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($publish_ip_address, $publish_web_presence, $authorization_status);
  ($packet, $authorization_status) = extract_byte($packet);
  ($packet, $publish_web_presence) = extract_byte($packet);
  ($packet, $publish_ip_address)   = extract_byte($packet);


  $DEBUG && warn("    authorization_status=$authorization_status\n",
                 "    publish_ip_address=$publish_ip_address\n",
                 "    publish_web_presence=$publish_web_presence\n");
                 

  my($user_info) = $self->db_user_info($user_uin);
  
  if (defined($user_info))
  {
    $$user_info{'publish_ip_address'}   = $publish_ip_address;
    $$user_info{'publish_web_presence'} = $publish_web_presence;  
    $$user_info{'authorization'}        = $authorization_status;

    $self->db_update_user_info($user_uin, $user_info);

    $packet = "";
    $packet =
      insert_uword($packet,
                   $server_meta_commands{'META_SRV_SET_SECURITY_REPLY'});
    $packet = insert_byte ($packet, 0x0A);
    
    $self->send_packet($self->construct_packet("META_USER_REPLY", $packet));
  }
  else
  {
    print "INFO_REQ: Don't know what to send\n";
  }
}

sub receive_meta_info_set_password
{
  my $self     = shift;
  my $packet   = shift;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($password);
  ($packet, $password) = extract_string($packet);


  $DEBUG && warn("    password=$password\n");

  my($user_info) = $self->db_user_info($user_uin);
  
  if (defined($user_info))
  {
    $$user_info{'password'} = crypt($password, $$user_info{'nick_name'});
    
    $self->db_update_user_info($user_uin, $user_info);

    $packet = "";
    $packet = 
      insert_uword($packet, 
                   $server_meta_commands{'META_SRV_SET_PASSWORD_REPLY'});
    $packet = insert_byte ($packet, 0x0A);
    
    $self->send_packet($self->construct_packet("META_USER_REPLY", $packet));
  }
  else
  {
    print "INFO_REQ: Don't know what to send\n";
  }
}

sub receive_meta_info_unknown1
{
  my $self     = shift;
  my $packet   = shift;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($searched_uin);
  ($packet, $searched_uin) = extract_udword($packet);


  $DEBUG && warn("    searched_uin=$searched_uin\n");

  print "Return list of offline messages here?\n";
  
  $self->send_packet($self->construct_packet("END_OFFLINE_MESSAGES"));
}

sub construct_meta_info_main
{
  my $self         = shift;
  my $searched_uin = shift;
  my $packet       = "";
  
  my($user_info) = $self->db_user_info($searched_uin);
  
  if (defined($user_info))
  {
    $packet = insert_uword($packet,
                           $server_meta_commands{'META_SRV_INFO_MAIN'});
    $packet = insert_byte ($packet, 0x0A);
    
    $packet = insert_string($packet, $$user_info{'nick_name'});
    $packet = insert_string($packet, $$user_info{'first_name'});
    $packet = insert_string($packet, $$user_info{'last_name'});
    $packet = insert_string($packet, $$user_info{'email'});
    $packet = insert_string($packet, $$user_info{'secondary_email'});
    $packet = insert_string($packet, $$user_info{'old_email'});
    $packet = insert_string($packet, $$user_info{'home_city'});
    $packet = insert_string($packet, $$user_info{'home_state'});
    $packet = insert_string($packet, $$user_info{'home_phone'});
    $packet = insert_string($packet, $$user_info{'home_fax'});
    $packet = insert_string($packet, $$user_info{'home_street'});
    $packet = insert_string($packet, $$user_info{'home_cellular'});
    
    $packet = insert_udword($packet, $$user_info{'home_zip_code'});
    $packet = insert_uword ($packet, $$user_info{'home_country_code'});
    $packet = insert_uword ($packet, $$user_info{'timezone'});
    $packet = insert_byte  ($packet, $$user_info{'publish_email_address'});
    $packet = insert_byte  ($packet, 0x00);
    $packet = insert_byte  ($packet, 0x00);
    
    return $self->construct_packet("META_USER_REPLY", $packet);  
  }
  else
  {
    warn "INFO_REQ: Don't know what to send";
    return "";
  }
}

sub construct_meta_info_work
{
  my $self         = shift;
  my $searched_uin = shift;
  my $packet       = "";
  
  my($user_info) = $self->db_user_info($searched_uin);
  
  if (defined($user_info))
  {
    $packet = insert_uword($packet,
                           $server_meta_commands{'META_SRV_INFO_WORK'});
    $packet = insert_byte ($packet, 0x0A);
    
    $packet = insert_string($packet, $$user_info{'work_city'});
    $packet = insert_string($packet, $$user_info{'work_state'});
    $packet = insert_string($packet, $$user_info{'work_phone'});
    $packet = insert_string($packet, $$user_info{'work_fax'});
    $packet = insert_string($packet, $$user_info{'work_street'});
    $packet = insert_udword($packet, $$user_info{'work_zip_code'});
    $packet = insert_uword ($packet, $$user_info{'work_country_code'});
    $packet = insert_string($packet, $$user_info{'work_company'});
    $packet = insert_string($packet, $$user_info{'work_division'});
    $packet = insert_string($packet, $$user_info{'work_position'});  
    $packet = insert_uword ($packet, $$user_info{'work_occupation'});
    $packet = insert_string($packet, $$user_info{'work_homepage'});
    
    return $self->construct_packet("META_USER_REPLY", $packet);  
  }
  else
  {
    warn "INFO_REQ: Don't know what to send";
    return "";
  }
}

sub construct_meta_info_about
{
  my $self         = shift;
  my $searched_uin = shift;
  my $packet       = "";
  
  my($user_info) = $self->db_user_info($searched_uin);
  
  if (defined($user_info))
  {
    $packet = insert_uword($packet,
                           $server_meta_commands{'META_SRV_INFO_ABOUT'});
    $packet = insert_byte ($packet, 0x0A);
    
    $packet = insert_string($packet, $$user_info{'about'});
    
    return $self->construct_packet("META_USER_REPLY", $packet);  
  }
  else
  {
    warn "INFO_REQ: Don't know what to send";
    return "";
  }
}

sub construct_meta_info_more
{
  my $self         = shift;
  my $searched_uin = shift;
  my $packet       = "";
  
  my($user_info) = $self->db_user_info($searched_uin);
  
  if (defined($user_info))
  {
    $packet = insert_uword ($packet,
                            $server_meta_commands{'META_SRV_INFO_MORE'});
    $packet = insert_byte  ($packet, 0x0A);
    
    $packet = insert_uword ($packet, $$user_info{'age'});
    $packet = insert_byte  ($packet, $$user_info{'sex'});
    $packet = insert_string($packet, $$user_info{'homepage'});
    $packet = insert_byte  ($packet, $$user_info{'birth_year'});
    $packet = insert_byte  ($packet, $$user_info{'birth_month'});
    $packet = insert_byte  ($packet, $$user_info{'birth_day'});
    $packet = insert_byte  ($packet, $$user_info{'language_1'});
    $packet = insert_byte  ($packet, $$user_info{'language_2'});
    $packet = insert_byte  ($packet, $$user_info{'language_3'});
    
    return $self->construct_packet("META_USER_REPLY", $packet);  
  }
  else
  {
    warn "INFO_REQ: Don't know what to send";
    return "";
  }
}

sub construct_meta_info_unknown_2
{
  my $self         = shift;
  my $searched_uin = shift;
  my $packet       = "";
  
  my($user_info) = $self->db_user_info($searched_uin);
  
  if (defined($user_info))
  {
    $packet =
      insert_uword($packet,
                   $server_meta_commands{'META_SRV_INFO_HOMEPAGE_CATEGORY'});
    $packet = insert_byte ($packet, 0x0A);
    
    $packet = insert_byte  ($packet, $$user_info{'category_1'});
    $packet = insert_byte  ($packet, $$user_info{'category_2'});
    $packet = insert_byte  ($packet, $$user_info{'category_3'});
    $packet = insert_string($packet, $$user_info{'category_text'});
    
    return $self->construct_packet("META_USER_REPLY", $packet);  
  }
  else
  {
    warn "INFO_REQ: Don't know what to send";
    return "";
  }
}

sub construct_meta_info_interests
{
  my $self         = shift;
  my $searched_uin = shift;
  my $packet       = "";
  
  my($user_info) = $self->db_user_info($searched_uin);
  
  if (defined($user_info))
  {
    $packet = insert_uword($packet,
                           $server_meta_commands{'META_SRV_INFO_INTERESTS'});
    $packet = insert_byte ($packet, 0x0A);
    
    $packet = insert_byte  ($packet, 4);
    $packet = insert_uword ($packet, $$user_info{'interest_type_1'});
    $packet = insert_string($packet, $$user_info{'interest_text_1'});
    $packet = insert_uword ($packet, $$user_info{'interest_type_2'});
    $packet = insert_string($packet, $$user_info{'interest_text_2'});
    $packet = insert_uword ($packet, $$user_info{'interest_type_3'});
    $packet = insert_string($packet, $$user_info{'interest_text_3'});
    $packet = insert_uword ($packet, $$user_info{'interest_type_4'});
    $packet = insert_string($packet, $$user_info{'interest_text_4'});
    
    return $self->construct_packet("META_USER_REPLY", $packet);  
  }
  else
  {
    warn "INFO_REQ: Don't know what to send";
    return "";
  }
}

sub construct_meta_info_affiliations
{
  my $self         = shift;
  my $searched_uin = shift;
  my $packet       = "";
  
  my($user_info) = $self->db_user_info($searched_uin);
  
  if (defined($user_info))
  {
    $packet =
      insert_uword($packet,
		   $server_meta_commands{'META_SRV_INFO_AFFILIATIONS'});
    $packet = insert_byte ($packet, 0x0A);
    
    $packet = insert_byte  ($packet, 3);
    $packet = insert_uword ($packet, $$user_info{'background_type_1'});
    $packet = insert_string($packet, $$user_info{'background_text_1'});
    $packet = insert_uword ($packet, $$user_info{'background_type_2'});
    $packet = insert_string($packet, $$user_info{'background_text_2'});
    $packet = insert_uword ($packet, $$user_info{'background_type_3'});
    $packet = insert_string($packet, $$user_info{'background_text_3'});
    $packet = insert_byte  ($packet, 3);
    $packet = insert_uword ($packet, $$user_info{'organization_type_1'});
    $packet = insert_string($packet, $$user_info{'organization_text_1'});
    $packet = insert_uword ($packet, $$user_info{'organization_type_2'});
    $packet = insert_string($packet, $$user_info{'organization_text_2'});
    $packet = insert_uword ($packet, $$user_info{'organization_type_3'});
    $packet = insert_string($packet, $$user_info{'organization_text_3'});
    
    return $self->construct_packet("META_USER_REPLY", $packet);  
  }
  else
  {
    warn "INFO_REQ: Don't know what to send";
    return "";
  }
}

sub receive_meta_info_get_unknown1
{
  my $self     = shift;
  my $packet   = shift;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($searched_uin);
  ($packet, $searched_uin) = extract_udword($packet);


  $DEBUG && warn("    searched_uin=$searched_uin\n");

  my($user_info) = $self->db_user_info($searched_uin);
  
  if (defined($user_info))
  {
    my $packet1 = $self->construct_meta_info_main($searched_uin);
    my $packet2 = $self->construct_meta_info_more($searched_uin);
    my $packet3 = $self->construct_meta_info_unknown_2($searched_uin);
    my $packet4 = $self->construct_meta_info_work($searched_uin);
    my $packet5 = $self->construct_meta_info_about($searched_uin);
    my $packet6 = $self->construct_meta_info_interests($searched_uin);
    my $packet7 = $self->construct_meta_info_affiliations($searched_uin);

    $self->send_multi_packet($packet1, $packet2, $packet3, $packet4,
                             $packet5, $packet6, $packet7);
  }
  else
  {
    print "INFO_REQ: Don't know what to send\n";
  }
}

sub receive_meta_info_all
{
  my $self     = shift;
  my $packet   = shift;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($searched_uin);
  ($packet, $searched_uin) = extract_udword($packet);


  $DEBUG && warn("    searched_uin=$searched_uin\n");

  my($user_info) = $self->db_user_info($searched_uin);
  
  if (defined($user_info))
  {
    my $packet1 = $self->construct_meta_info_main($user_uin);  
    my $packet2 = $self->construct_meta_info_more($user_uin);
    my $packet3 = $self->construct_meta_info_unknown_2($user_uin);
    my $packet4 = $self->construct_meta_info_work($user_uin);
    my $packet5 = $self->construct_meta_info_about($user_uin);  
    my $packet6 = $self->construct_meta_info_interests($user_uin);
    my $packet7 = $self->construct_meta_info_affiliations($user_uin);
    my $packet8 = 
      $self->construct_packet("END_OFFLINE_MESSAGES",
                              "",
                              $self->{'sessions'}{$user_uin}{'sequence_number'}++,
                              0);
    $self->send_multi_packet($packet1, $packet2, $packet3, $packet4,
                             $packet5, $packet6, $packet7, $packet8);
  }
  else
  {
    print "INFO_REQ: Don't know what to send\n";
  }
}

sub receive_meta_info_search_uin
{
  my $self     = shift;
  my $packet   = shift;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($searched_uin);
  ($packet, $searched_uin) = extract_udword($packet);


  $DEBUG && warn("    searched_uin=$searched_uin\n");

  
  my($user_info) = $self->db_user_info($searched_uin);
  
  if (defined($user_info))
  {
    $self->send_meta_info_search_reply(0, $searched_uin); 
  }
  else
  {
    $self->send_meta_info_search_reply(0, ()); 
  }
}

sub receive_meta_info_search_email
{
  my $self     = shift;
  my $packet   = shift;
  my $wildcard = shift || 0;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($searched_email);
  ($packet, $searched_email) = extract_string($packet);


  $DEBUG && warn("    searched_email=$searched_email\n",
                 "    wildcard=$wildcard\n");


  $searched_email  = "*" if ($searched_email  eq "");
  
  $searched_email  =~ s/\*/%/g;
  
  
  my(@keys) = ();
  
  my $sth =
    $self->db_do("SELECT uin FROM users WHERE " .
                 "(email           LIKE '$searched_email') OR " .
                 "(secondary_email LIKE '$searched_email') OR " .
                 "(old_email       LIKE '$searched_email')" .
                 ";");
  while (my $pointer = $sth->fetchrow_hashref)
  {
    push(@keys, $pointer->{'uin'});
  }

  $self->send_meta_info_search_reply(0, @keys); 
}

sub receive_meta_info_search_email_wildcard
{
  my $self     = shift;
  my $packet   = shift;

  # Just call the normal version
  $self->receive_meta_info_search_email($packet, 1);
}

sub receive_meta_info_search_details
{
  my $self     = shift;
  my $packet   = shift;
  my $wildcard = shift || 0;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($searched_nick, $searched_first, $searched_last);
  ($packet, $searched_first) = extract_string($packet);
  ($packet, $searched_last)  = extract_string($packet);
  ($packet, $searched_nick)  = extract_string($packet);


  $DEBUG && warn("    searched_nick=$searched_nick\n",
                 "    searched_first=$searched_first\n",
                 "    searched_last=$searched_last\n",
                 "    wildcard=$wildcard\n");

  $searched_nick  = "*" if ($searched_nick  eq "");
  $searched_first = "*" if ($searched_first eq "");
  $searched_last  = "*" if ($searched_last  eq "");
  
  $searched_nick  =~ s/\*/%/g;
  $searched_first =~ s/\*/%/g;
  $searched_last  =~ s/\*/%/g;
  
  
  my(@keys) = ();
  
  my $sth =
    $self->db_do("SELECT uin FROM users WHERE " .
                 "(nick_name  LIKE '$searched_nick') AND " .
                 "(first_name LIKE '$searched_first') AND " .
                 "(last_name  LIKE '$searched_last')" .
                 ";");
  while (my $pointer = $sth->fetchrow_hashref)
  {
    push(@keys, $pointer->{'uin'});
  }
  
  $self->send_meta_info_search_reply(0, @keys); 
}

sub receive_meta_info_search_details_wildcard
{
  my $self     = shift;
  my $packet   = shift;

  # Just call the normal version
  $self->receive_meta_info_search_details($packet, 1);
}

sub receive_meta_info_search_whitepages
{
  my $self     = shift;
  my $packet   = shift;
  my $wildcard = shift || 0;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($nick_name, $first_name, $last_name, $email, $age_start, $age_end, $sex,
     $language, $city, $state, $country_code, $work_company, $work_division,
     $work_position, $work_occupation, $background_type, $background_text,
     $interest_type, $interest_text, $organization_type, $organization_text,
     $show_only_online);

  ($packet, $first_name)            = extract_string($packet);
  ($packet, $last_name)             = extract_string($packet);
  ($packet, $nick_name)             = extract_string($packet);
  ($packet, $email)                 = extract_string($packet);
  ($packet, $age_start)             = extract_uword ($packet);
  ($packet, $age_end)               = extract_uword ($packet);
  ($packet, $sex)                   = extract_byte  ($packet);
  ($packet, $language)              = extract_byte  ($packet);
  ($packet, $city)                  = extract_string($packet);
  ($packet, $state)                 = extract_string($packet);
  ($packet, $country_code)          = extract_uword ($packet);
  ($packet, $work_company)          = extract_string($packet);
  ($packet, $work_division)         = extract_string($packet);
  ($packet, $work_position)         = extract_string($packet);
  ($packet, $work_occupation)       = extract_byte  ($packet); # Word?
  ($packet, $background_type)       = extract_uword ($packet);
  ($packet, $background_text)       = extract_string($packet);
  ($packet, $interest_type)         = extract_uword ($packet);
  ($packet, $interest_text)         = extract_string($packet);
  ($packet, $organization_type)     = extract_uword ($packet);
  ($packet, $organization_text)     = extract_string($packet);
  ($packet, $show_only_online)      = extract_byte  ($packet);


  $DEBUG && warn("    first_name=$first_name\n",
                 "    last_name=$last_name\n",
                 "    nick_name=$nick_name\n",
                 "    email=$email\n",
                 "    age_start=$age_start\n",
                 "    age_end=$age_end\n",
                 "    sex=$sex\n",
                 "    language=$language\n",
                 "    city=$city\n",
                 "    state=$state\n",
                 "    country_code=$country_code\n",
                 "    work_company=$work_company\n",
                 "    work_division=$work_division\n",
                 "    work_position=$work_position\n",
                 "    work_occupation=$work_occupation\n",
                 "    background_type=$background_type\n",
                 "    background_text=$background_text\n",
                 "    interest_type=$interest_type\n",
                 "    interest_text=$interest_text\n",
                 "    organization_type=$organization_type\n",
                 "    organization_text=$organization_text\n",
                 "    show_only_online=$show_only_online\n",
                 "    wildcard=$wildcard\n");

  my($SQL) =
    "SELECT uin FROM users WHERE " .
    ($nick_name         ?
     "(nick_name            LIKE '$nick_name') AND "          : "") .
    ($first_name        ?
     "(first_name           LIKE '$first_name') AND "         : "") .
    ($last_name         ?
     "(last_name            LIKE '$last_name') AND "          : "") .
    ($email             ?
     "((email               LIKE '$email') OR " .
     " (secondary_email     LIKE '$email') OR " .
     " (old_email           LIKE '$email')) AND "             : "") .
    ($age_start && $age_start != 65535 ?
     "(age                  >=   '$age_start') AND "          : "") .
    ($age_end && $age_end != 65535 ?
     "(age                  <=   '$age_end') AND "            : "") .
    ($sex               ?
     "(sex                  =    '$sex') AND "                : "") .
    ($language && $language != 255 ?
     "((language_1          =    '$language') OR " .
     " (language_2          =    '$language') OR " .
     " (language_3          =    '$language')) AND "          : "") .
    ($city              ?
     "((home_city           LIKE '$city') OR " .
     " (work_city           LIKE '$city')) AND "              : "") .
    ($state             ?
     "((home_state          LIKE '$state') OR " .
     " (work_state          LIKE '$state')) AND "             : "") .
    ($country_code && $country_code != 65535 ?
     "((home_country_code   LIKE '$country_code') OR " .
     " (work_country_code   LIKE '$country_code')) AND "      : "") .
    ($work_company      ?
     "(work_company         LIKE '$work_company') AND "       : "") .
    ($work_division     ?
     "(work_division        LIKE '$work_division') AND "      : "") .
    ($work_position     ?
     "(work_position        LIKE '$work_position') AND "      : "") .
    ($work_occupation && $work_occupation != 255 ?
     "(work_occupation      =    '$work_occupation') AND "    : "") .
    ($background_type   ?
     "((background_type_1   = '$background_type') OR " .
     " (background_type_2   = '$background_type') OR " .
     " (background_type_3   = '$background_type')) AND "      : "") .
#   ($background_text   ?
#    "(background_text      LIKE '$background_text') AND "    : "") .
    ($interest_type     ?
     "((interest_type_1     LIKE '$interest_type') OR " .
     " (interest_type_2     LIKE '$interest_type') OR " .
     " (interest_type_3     LIKE '$interest_type') OR " .
     " (interest_type_4     LIKE '$interest_type')) AND "     : "") .
#   ($interest_text     ?
#    "(interest_text     LIKE '$interest_text') AND "         : "") .
    ($organization_type ?
     "((organization_type_1 LIKE '$organization_type') OR " .
     " (organization_type_2 LIKE '$organization_type') OR " .
     " (organization_type_3 LIKE '$organization_type')) AND " : "") .
#   ($organization_text ?
#    "(organization_text LIKE '$organization_text') AND "     : "") .
    ($show_only_online  ?
     "(status               !=   '$status_bits{'OFFLINE'}') AND " .
     "(status               &    '$status_bits{'INVISIBLE'}' = 0) AND " : "") .
    "1;";
  $SQL =~ s/\*/%/g;

  my $sth = $self->db_do($SQL);
  my(@keys) = ();
  while (my $pointer = $sth->fetchrow_hashref)
  {
    push(@keys, $pointer->{'uin'});
  }

  $self->send_meta_info_search_reply(1, @keys); 
}

sub receive_meta_info_search_whitepages_wildcard
{
  my $self     = shift;
  my $packet   = shift;

  # Just call the normal version
  $self->receive_meta_info_search_whitepages($packet, 1);
}

# META_INFO_UNKNOWN2 is received right after LOGIN sometimes...

# Incoming decrypted packet:
# 0000: 05 00 00 00 00 00 A2 87 01 00 26 53 C3 18 4A 06 .........&S.J.
# 0010: 60 24 02 00 3E 38 EA 17 B8 06 E8 00 D7 00 A8 C1 `$..>8....
# 0020: 30 37 00 00 00 00 00 00 00 00 00 00 00 00 00 00 07..............
# 0030: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 ................
# 0040: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 ................
# 0050: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 ................
# 0060: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 ................
# 0070: 00 00 00 00 C2 38 37 37 F7 09 00 00 F7 09 00 00 ....877......
# 0080: 00 00 00 00 2C 06 00 00 00 00 00 00 2C 06 00 00 ....,.......,...
# 0090: E5 09 00 00 E5 09 00 00 00 00 00 00 1B 06 00 00 ..............
# 00A0: 00 00 00 00 1B 06 00 00 03 09 00 00 03 09 00 00 ................
# 00B0: 00 00 00 00 3E 05 00 00 00 00 00 00 3E 05 00 00 ....>.......>...
# 00C0: 03 09 00 00 03 09 00 00 00 00 00 00 3E 05 00 00 ............>...
# 00D0: 00 00 00 00 3E 05 00 00 01 00 C2 38 37 37 C2 38 ....>.....8778
# 00E0: 37 37 01 00 BC 09 00 00 00 85 00 00 00 00 00 00 77.............
# 00F0: 00 40 B2 00 77 01 DF 04 D1 03 20 00 00 00 00 00 .@.w... .....
# 0100: 00 90 F6 07                                     ...            
# INCOMING:  Got command META_USER (1610/0x064A) sequence 9312, 2
# <<META_USER:
#     user_uin=100258
#     meta_cmd=META_INFO_UNKNOWN2 (0x06B8)
# ##### Not implemented receive_meta_info_unknown2()!!
# construct_packet:  command 'META_USER_REPLY', seq_num '1, 2'

sub receive_meta_info_unknown2
{
  my $self     = shift;
  my $packet   = shift;
  my $version  = $self->{'packet_header'}->{'version'};
  my $user_uin = $self->{'packet_header'}->{'user_uin'};

  my($about);
# ($packet, $about) = extract_string($packet);

# $DEBUG && warn("    about=$about\n");

  print "##### Not implemented receive_meta_info_unknown2()!!\n";
  
  $packet = "";
  $packet = insert_uword($packet, $server_meta_commands{'META_SRV_UNKNOWN_2'});
  $packet = insert_byte ($packet, 0x32);

  $self->send_packet($self->construct_packet("META_USER_REPLY", $packet));
}

sub receive_client_unknown_1
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin, $unknown);
  ($packet, $user_uin) = $self->extract_or_get_user_uin($packet);
  ($packet, $unknown)  = extract_udword($packet);


  $DEBUG && warn("<<CLIENT_UNKNOWN_1:\n",
                 "    user_uin=$user_uin\n",
                 "    unknown=$unknown\n");

  $self->{'user_uin'} = $user_uin;
}

sub receive_client_unknown_2
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin);
  ($packet, $user_uin) = $self->extract_or_get_user_uin($packet);


  $DEBUG && warn("<<CLIENT_UNKNOWN_2:\n",
                 "    user_uin=$user_uin\n");

  $self->{'user_uin'} = $user_uin;
}

sub receive_client_unknown_3
{
  my $self    = shift;
  my $packet  = shift;
  my $version = $self->{'packet_header'}->{'version'};

  my($user_uin);
  ($packet, $user_uin) = $self->extract_or_get_user_uin($packet);


  $DEBUG && warn("<<CLIENT_UNKNOWN_3:\n",
                 "    user_uin=$user_uin\n");

  $self->{'user_uin'} = $user_uin;
}


###################################################################
# Database Functions                                              #
###################################################################

sub db_init
{
  my $self = shift;
  my $dbh  =
    DBI->connect("DBI:mysql:database=$config::database_name;host=$config::database_host",
                 $config::database_username,
                 $config::database_password);
  die "Cannot connect: $DBI::errstr\n" unless $dbh;
  $self->{'database'} = $dbh;

  print "SQL database connected\n";
  
  $self->db_create_tables;
  
  if (0)
  {
    $self->db_print_database;
  }
}

sub db_close
{
  my $self = shift;
  
  my $dbh = $self->{'database'};

  $dbh->disconnect;
}


sub db_do
{
  my $self  = shift;
  my $SQL   = shift;
  my $dbh   = $self->{'database'};
  my $sth;

  eval
  {
    $sth = $dbh->prepare($SQL);
  };
  
   # Check for errors.
  if ($@)
  {
    $dbh->disconnect;
    die "SQL error: $@\n";
  } 
  else 
  {
    $sth->execute;
  }
  
  return ($sth);
}

sub db_create_tables
{
  my $self = shift;
  
  my $sth = $self->db_do("SHOW TABLES;");
  my %tables = ();
  while (my $pointer = $sth->fetchrow_array)
  {
    $tables{$pointer} = 1;
  }
  
  if (!defined($tables{'users'}))
  {
    print "Create users table:\n";

    my @user_table =
    (
      "uin                   INTEGER       NOT NULL",
      "password              VARCHAR(13)   NOT NULL",
      "nick_name             VARCHAR(20)   NOT NULL",
      "first_name            VARCHAR(20)   NOT NULL",
      "last_name             VARCHAR(20)   NOT NULL",
      "email                 VARCHAR(60)   NOT NULL",
      "secondary_email       VARCHAR(60)   NOT NULL",
      "old_email             VARCHAR(60)   NOT NULL",

      "authorization         INTEGER       NOT NULL",
      "publish_web_presence  INTEGER       NOT NULL",
      "publish_ip_address    INTEGER       NOT NULL",
      "publish_email_address INTEGER       NOT NULL",

      "home_street           VARCHAR(30)   NOT NULL",
      "home_city             VARCHAR(30)   NOT NULL",
      "home_state            VARCHAR(3)    NOT NULL",
      "home_zip_code         INTEGER       UNSIGNED NOT NULL",
      "home_country_code     INTEGER       NOT NULL",
      "home_phone            VARCHAR(30)   NOT NULL",
      "home_fax              VARCHAR(30)   NOT NULL",
      "home_cellular         VARCHAR(30)   NOT NULL",

      "timezone              INTEGER       NOT NULL",

      "category_1            INTEGER       NOT NULL",
      "category_2            INTEGER       NOT NULL",
      "category_3            INTEGER       NOT NULL",
      "category_text         TEXT          NOT NULL",

      "age                   INTEGER       NOT NULL",
      "sex                   INTEGER       NOT NULL",
      "homepage              VARCHAR(100)  NOT NULL",
      "birth_year            INTEGER       NOT NULL",
      "birth_month           INTEGER       NOT NULL",
      "birth_day             INTEGER       NOT NULL",
      "language_1            INTEGER       NOT NULL",
      "language_2            INTEGER       NOT NULL",
      "language_3            INTEGER       NOT NULL",

      "work_company          VARCHAR(30)   NOT NULL",
      "work_division         VARCHAR(30)   NOT NULL",
      "work_position         VARCHAR(30)   NOT NULL",
      "work_occupation       INTEGER       NOT NULL",
      "work_phone            VARCHAR(30)   NOT NULL",
      "work_fax              VARCHAR(30)   NOT NULL",
      "work_street           VARCHAR(30)   NOT NULL",
      "work_city             VARCHAR(30)   NOT NULL",
      "work_state            VARCHAR(3)    NOT NULL",
      "work_zip_code         INTEGER       UNSIGNED NOT NULL",
      "work_country_code     INTEGER       NOT NULL",
      "work_homepage         VARCHAR(100)  NOT NULL",
 
      "organization_type_1   INTEGER       NOT NULL",
      "organization_text_1   VARCHAR(60)   NOT NULL",
      "organization_type_2   INTEGER       NOT NULL",
      "organization_text_2   VARCHAR(60)   NOT NULL",
      "organization_type_3   INTEGER       NOT NULL",
      "organization_text_3   VARCHAR(60)   NOT NULL",
 
      "background_type_1     INTEGER       NOT NULL",
      "background_text_1     VARCHAR(60)   NOT NULL",
      "background_type_2     INTEGER       NOT NULL",
      "background_text_2     VARCHAR(60)   NOT NULL",
      "background_type_3     INTEGER       NOT NULL",
      "background_text_3     VARCHAR(60)   NOT NULL",

      "interest_type_1       INTEGER       NOT NULL",
      "interest_text_1       VARCHAR(60)   NOT NULL",
      "interest_type_2       INTEGER       NOT NULL",
      "interest_text_2       VARCHAR(60)   NOT NULL",
      "interest_type_3       INTEGER       NOT NULL",
      "interest_text_3       VARCHAR(60)   NOT NULL",
      "interest_type_4       INTEGER       NOT NULL",
      "interest_text_4       VARCHAR(60)   NOT NULL",

      "about                 TEXT          NULL",
     
      "status                INTEGER       UNSIGNED NOT NULL",
      "login_time            DATETIME      NOT NULL",
    );
    
    my $SQL = 
      "CREATE TABLE users (\n  " .
      join("\n, ", @user_table) . "\n, PRIMARY KEY uin (uin))";
    
    print "$SQL\n";
    $self->db_do($SQL);
  }
  
  if (!defined($tables{'messages'}))
  {
    print "Create messages table:\n";

    my @messages_table = 
    (
      "uin                   INTEGER       NOT NULL",
      "sender_uin            INTEGER       NOT NULL",
      "message_type          VARCHAR(20)   NOT NULL",
      "message               VARCHAR(255)  NOT NULL",
      "timestamp             TIMESTAMP(14) NOT NULL",
    );
    
    my $SQL = 
      "CREATE TABLE messages (\n  " .
      join("\n, ", @messages_table) . ")";
    
    print "$SQL\n";
    $self->db_do($SQL);
  }
  
#  if (defined($tables{'sessions'}))
  {
    $self->db_do("UPDATE users SET status = '" . $status_bits{'OFFLINE'} . 
                 "';");
  }
  
  if (0)
  {
    print "Create sessions table:\n";
    
    my @sessions_table = 
    (
      "uin               INTEGER       NOT NULL",
      "status            VARCHAR(20)   NOT NULL",
      "login_time        DATETIME      NOT NULL",
    );
    
    my $SQL = 
      "CREATE TABLE sessions (\n  " .
      join("\n, ", @sessions_table) . "\n, PRIMARY KEY uin (uin))";
    
    print "$SQL\n";
    $self->db_do($SQL);
  }
  
#  print "Try to alter table\n";
#  if (0)
#    {
#      $SQL =
#       "ALTER TABLE $user_table ADD (full_name char(80) NULL);";
#      $self->db_do($SQL);
#    }
}

sub db_print_database
{
  my $self = shift;
  my $sth;

  if (0)
  {
    print "Columns:\n";  
    $sth = $self->db_do("SHOW COLUMNS FROM users;");
    while (my $pointer = $sth->fetchrow_hashref)
    {
      my ($field) = $pointer->{'Field'};
      
      print "$field ";
    }
    print "\n";
  }
  

  print "\nTable contents:\n";

  $sth = 
    $self->db_do("SELECT uin, nick_name, first_name, last_name FROM users;");
  while (my $pointer = $sth->fetchrow_hashref)
  {
    my ($uin, $first_name, $last_name, $nick_name) =
      ($pointer->{'uin'},
       $pointer->{'first_name'},
       $pointer->{'last_name'},
       $pointer->{'nick_name'});
    
    print 
      "UIN  : $uin\n", 
      "Nick : $nick_name\n",
      "First: $first_name\n",
      "Last : $last_name\n", 
      "\n";
  }
  $sth->finish;
}

sub db_all_uins
{
  my $self = shift;
  my $sth;
  my @uins = ();

  $sth = $self->db_do("SELECT uin FROM users");
  while (my $pointer = $sth->fetchrow_hashref)
  {
    my $uin = $pointer->{'uin'};

    push(@uins, $uin);
  }
  $sth->finish;
  
  return sort(@uins);
}

sub db_exact_search_by_name
{
  my $self       = shift;
  my $nick_name  = uc(shift);
  my $first_name = uc(shift);
  my $last_name  = uc(shift);

  my @keys = ();
  my $sth = 
    $self->db_do("SELECT uin, nick_name, first_name, last_name FROM users");
  while (my $pointer = $sth->fetchrow_hashref)
  {
    if ((!length($nick_name) ||
         uc($pointer->{'nick_name'}) eq $nick_name) &&
        (!length($first_name) ||
         uc($pointer->{'first_name'}) eq $first_name) &&
        (!length($last_name) ||
         uc($pointer->{'last_name'}) eq $last_name))
    {
      if (scalar(@keys) == $max_search_hits)
      {
        $sth->finish;
        return(1, @keys);
      }
      
      push(@keys, $pointer->{'uin'});
    }
  }
  
  $sth->finish;
  return (0, @keys);
}

sub db_exact_search_by_email
{
  my $self  = shift;
  my $email = uc(shift);

  my @keys = ();
  my $sth = $self->db_do("SELECT uin, email FROM users");
  while (my $pointer = $sth->fetchrow_hashref)
  {
    if (!length($email) ||
        uc($pointer->{'email'}) eq $email)
    {
      if (scalar(@keys) == $max_search_hits)
      {
        $sth->finish;
        return(1, @keys);
      }
      
      push(@keys, $pointer->{'uin'});
    }
  }
  
  $sth->finish;
  return (0, @keys);
}

sub db_user_info
{
  my $self     = shift;
  my $user_uin = shift;

  my $sth = 
    $self->db_do("SELECT * FROM users WHERE uin = $user_uin");
  if (my $pointer = $sth->fetchrow_hashref)
  {
    return $pointer;
  }
  
  return undef;
}


sub db_create_new_user
{
  my $self     = shift;
  my $user_uin = shift;

  my $uid = $user_uin;
  # Shift down the uid until we get below the base_uin
  while ($uid > $config::base_uin)
  {
    $uid -= $config::base_uin;
  }

  my($user_name, $gecos) = (getpwuid($uid))[0, 6];
  
  my($full_name) = (split(',', $gecos))[0];
  my($first_name, $last_name) = split(' ', $full_name);

  my $SQL =
    "INSERT INTO users " .
    "(uin, nick_name, first_name, last_name, email, homepage, " .
    "home_country_code, timezone, home_city, home_phone, authorization, " .
    "about, status, category_text, " .
    "work_company, work_street, work_city, work_zip_code, " . 
    "work_country_code, work_homepage, work_phone, work_fax" .
    ") " .
    "VALUES ($user_uin, '$first_name $last_name', '$first_name', " .
    "'$last_name', '$user_name\@$config::default_email_domain_name', " .
    "'$config::default_homepage_prefix$user_name/', " .
    "'$config::default_country_code', '$config::default_timezone', " .
    "'$config::default_city', '$config::default_phone', " .
    "'$config::default_authorization', '', '$status_bits{'OFFLINE'}', '', " .
    "'$config::default_work_company', '$config::default_work_street', " .
    "'$config::default_work_city', '$config::default_work_zip_code', " .
    "'$config::default_work_country_code', " .
    "'$config::default_work_homepage_prefix$user_name/', " .
    "'$config::default_work_phone', '$config::default_work_fax'" .
    ");";
  print "$SQL\n";
  $self->db_do($SQL);
}


sub db_update_user_info
{
  my $self      = shift;
  my $user_uin  = shift;
  my $user_info = shift;
  
  my $SQL = "UPDATE users SET ";
  my $first = 1;
  foreach my $field (keys(%{$user_info}))
  {
    $SQL .= ", " if (!$first);
    $first = 0;
    
    $SQL .= "$field = '$user_info->{$field}' ";    
  }

  $SQL .= " WHERE uin = $user_uin;";
  
  $self->db_do($SQL);
}


sub session_set_user_status
{
  my $self     = shift;
  my $user_uin = shift;
  my $status   = shift;

  $self->{'sessions'}{$user_uin}{'status'} = $status;

  my($sec, $min, $hour, $mday, $month, $year, $wday, $yday, $isdst) =
    localtime(time);
  $year += 1900;
  $month++;
  
  my $was_offline = $self->session_check_user_status($user_uin, 'OFFLINE');

  if ($was_offline && $status != $status_bits{'OFFLINE'})
  {
    # Update last login time
    $self->db_do(sprintf("UPDATE users SET login_time = " .
                         "'%04d%02d%02d%02d%02d%02d' WHERE uin = $user_uin;",
                         $year, $month, $mday, $hour, $min, $sec));
  }
  
  
  # Change status
  $self->db_do("UPDATE users SET status = '$status' WHERE uin = $user_uin;");

  # Notify everyone who is interested in my status about the change
  foreach my $notify_uin (keys(%{$self->{'sessions'}{$user_uin}{'notified'}}))
  {
    $self->notify_user($user_uin, $notify_uin);
  }
}

sub session_get_user_status
{
  my $self     = shift;
  my $user_uin = shift;
  my $status;

  my $sth = $self->db_do("SELECT status FROM users WHERE uin = $user_uin;");
  if (my $pointer = $sth->fetchrow_hashref)
  {
    $status = $pointer->{'status'};
  }
  # All non-existing users are considered to be offline
  else
  {
    $status = $status_bits{'OFFLINE'};
  }
  $sth->finish;

  return $status;
}

sub session_check_user_status
{
  my $self     = shift;
  my $user_uin = shift;
  my $status   = $self->session_get_user_status($user_uin);

#print "session_check_user_status: $user_uin: ", udword_to_hex($status), "\n";
#print "args: ", join(", ", @_), "\n";

  if ($_[0] eq 'ONLINE')
  {
    return ($status & 0xFFFF & ~$status_bits{'FREE_CHAT'}) == $status_bits{'ONLINE'};
  }
  elsif ($_[0] eq 'OFFLINE')
  {
    return $status == $status_bits{'OFFLINE'};
  }

  foreach (@_)
  {
    return 0 unless ($status_bits{$_} & $status);
  }

  return 1;
}

%user_status =
(
  0x00000000 => "Online",
  0x00000001 => "Away",
# 0x00000002 => "Accept Urgent",
# 0x00000004 => "N/A",
  0x00000005 => "N/A (Extended Away)",
# 0x00000010 => "DND1",
  0x00000011 => "Occupied (Urgent Messages)",
  0x00000013 => "DND (Do Not Disturb)",
  0x00000020 => "Free for Chat",
  0x00000100 => "Privacy (Invisible)",
  0xFFFFFFFF => "Offline",
);

sub status_to_text
{
  my $self     = shift;
  my $status   = shift;

  my $result;
  
  if (defined($user_status{$status & 0xFFFF}))
  {
    $result = $user_status{$status & 0xFFFF};
  }
  else
  {
    return uword_to_hex($status);
  }

  $result .= ' Birthday' if ($status & $status_bits{'BIRTHDAY'});
  $result .= ' Unknown'  if ($status & $status_bits{'UNKNOWN'});

  return $result;
}

sub session_kill_inactive_sessions
{
  my $self = shift;
  
  foreach my $user_uin (keys(%{$self->{'sessions'}}))
  {
    if (!$self->session_check_user_status($user_uin, 'OFFLINE'))
    {
      my $old_time = 
	$self->{'sessions'}{$user_uin}{'last_packet_received_time'};
      
      # Kick out inactive sessions
      if ((time - $old_time) > $session_timeout)
      {
        $self->session_set_user_status($user_uin, $status_bits{'OFFLINE'});
      }
    }
  }
}


=head1 MISC FUNCTIONS

These do not correspond with anything much.

=item version ( );

Returns version information for this module.

=cut

sub version
{
  return "ICQServer version $VERSION";
}


###################################################################
# Functions under here are helper functions                       #
###################################################################

sub send_packet
{
  my $self   = shift;
  my $packet = shift;
  my $user_uin = $self->{'user_uin'};

  if (!defined($self->{'socket'}->send($packet)))
  {
    carp "send: $!";
    return 0;
  }

  $self->{'sessions'}{$user_uin}{'sequence_number2'}++ if ($user_uin);

  return 1;
}

sub construct_packet
{
  my $self     = shift;
  my $command  = shift;
  my $data     = shift || '';  # Assume data is already packed or whatever
  my $user_uin = $self->{'user_uin'};
  if ($self->session_check_user_status($user_uin, 'OFFLINE'))
  {
    $user_uin = 0;
  }
  
  my $seq_num1 =
    (@_ ? shift :
     ($user_uin ?
      $self->{'sessions'}{$user_uin}{'sequence_number'}++ :
      0));
  my $seq_num2 =
    (@_ ? shift :
     ($user_uin ?
      $self->{'sessions'}{$user_uin}{'sequence_number2'} :
      0));
  my $session_id = $user_uin ? 
    $self->{'sessions'}{$user_uin}{'session_id'} :
    $self->{'packet_header'}{'session_id'};
  
  $DEBUG && warn "construct_packet:  command '$command', seq_num '$seq_num1, $seq_num2'\n" if ($DEBUG_ACK || ($command ne "ACK"));

  my($packet);
  my($version) = $self->{'packet_header'}{'version'};

  if ($version == 2)
  {
    $packet = 
      pack("vvv",
           $version,
           $server_commands{$command},
           $seq_num1) .
      $data;

    if ($DEBUG && $DEBUG_PACKET_DUMP)
    {
      print "\nOutgoing packet:\n";
      $self->dump_packet($packet);
    }
  }
  elsif ($version == 3 || $version == 4)
  {
    $packet = 
      pack("vvvvVV",
           $version,
           $server_commands{$command},
           $seq_num1,
           $seq_num2,
           $self->{'user_uin'},
           0) .                  # check code
      $data;

    my($checkcode) = $self->calculate_checkcode_v5($packet);

    $packet = $self->insert_checkcode_v3_srv($packet, $checkcode);
    
    if ($DEBUG && $DEBUG_PACKET_DUMP &&
        ($DEBUG_ACK || $command ne "ACK"))
    {
      print "\nOutgoing packet:\n";
      $self->dump_packet($packet);
    }
  }
  elsif ($version == 5)
  {
    $packet = 
      pack("vcVvvvVV",
           $version,
           0,
           $session_id,
           $server_commands{$command},
           $seq_num1,
           $seq_num2,
           $self->{'user_uin'},
           0) .                  # check code
      $data;

    my($checkcode) = $self->calculate_checkcode_v5($packet);

    $packet = $self->insert_checkcode_v5_srv($packet, $checkcode);
    
    if ($DEBUG && $DEBUG_PACKET_DUMP &&
        ($DEBUG_ACK || $command ne "ACK"))
    {
      print "\nOutgoing packet:\n";
      $self->dump_packet($packet);
    }
  }
  else
  {
    die("Cannot handle version $version packets!!!\n");
  }

  return $packet;
}

sub notify_user
{
  my $self       = shift;
  my $user_uin   = shift;
  my $notify_uin = shift;

  if ($self->session_check_user_status($notify_uin, 'OFFLINE'))
  {
    return;
  }

  $self->{'user_uin'} = $notify_uin;

  my $notify_ip_address = $self->{'sessions'}{$notify_uin}{'ip_address'};
  print "Notifying $notify_ip_address\n";

  my $receiver =
    sockaddr_in($self->{'sessions'}{$notify_uin}{'udp_port'},
                inet_aton($notify_ip_address));
  
  if ($self->session_check_user_status($user_uin, 'OFFLINE'))
  {
    $self->{'socket'}->send($self->construct_packet("USER_OFFLINE",
                                                    pack("V", $user_uin)),
                            0, $receiver);
  }
  else
  {
    my $ip_address = $self->{'sessions'}{$user_uin}{'ip_address'};
    my $real_ip_address = $self->{'sessions'}{$user_uin}{'real_ip_address'};
    my $tcp_port = $self->{'sessions'}{$user_uin}{'tcp_port'};
    my $packet = "";
    $packet = insert_udword($packet, $user_uin);
    $packet = insert_ipaddr($packet, unpack("N", inet_aton($ip_address)));
    $packet = insert_udword($packet, $tcp_port);
    $packet = insert_ipaddr($packet, unpack("N", inet_aton($real_ip_address)));
    $packet = insert_byte  ($packet, 0x04);
    $packet = insert_udword($packet, $self->{'sessions'}{$user_uin}{'status'});
    $packet = insert_udword($packet, 0x00000200);
    
    $self->{'socket'}->send($self->construct_packet("USER_ONLINE", $packet),
                            0, $receiver);
  }
}

sub message_to_user
{
  my $self   = shift;
  my $packet = "";
  my($user_uin, $receiver_uin, $message_type, $message) = @_;

  my($user_info) = $self->db_user_info($receiver_uin);
  
  if (defined($user_info))
  {
    if (!$self->session_check_user_status($receiver_uin, 'OFFLINE'))
    {
      print "Send message for: $receiver_uin\n";
      my $receiver_ip = 
        inet_aton($self->{'sessions'}{$receiver_uin}{'ip_address'});
      my $receiver =
        sockaddr_in($self->{'sessions'}{$receiver_uin}{'udp_port'}, 
                    $receiver_ip);
      
      $self->{'user_uin'} = $receiver_uin;

      $packet = "";
      $packet = insert_udword($packet, $user_uin);
      $packet = insert_uword ($packet, $message_type);
      $packet = insert_string($packet, $message);
      $self->{'socket'}->
        send($self->construct_packet("SYSTEM_DELIVERED_MESSAGE", $packet),
             0, $receiver);
    }
    else
    {
      print "Save message for: $receiver_uin\n";
      $self->db_do("INSERT INTO messages " .
                   "(uin, sender_uin, message_type, message) VALUES " .
                   "('$receiver_uin', '$user_uin', " .
                   "'$message_type', '$message');");
    }
  }
  
  # We have to increase sequence_number2 even if we don't send back a
  # message to the sender
  $self->{'sessions'}{$user_uin}{'sequence_number2'}++;
}

sub verify_password
{
  my $self     = shift;
  my $user_uin = shift;
  my $password = shift;

  return 0 if ($user_uin < $config::base_uin);
  
  # If a password is set for the user, we use that
  my $sth = 
    $self->db_do("SELECT password FROM users WHERE uin = $user_uin;");
  if (my $pointer = $sth->fetchrow_hashref)
  {
    my $passwd = $pointer->{'password'};
    if (crypt($password, substr($passwd, 0, 2)) eq $passwd)
    {
      return 1;
    }
  }
  
  # No password set, authenticate against the Unix passwd database instead
  
  # Shift down the uid until we get below the base_uin
  my $uid = $user_uin;  
  while ($uid > $config::base_uin)
  {
    $uid -= $config::base_uin;
  }
  
  my $passwd = (getpwuid($uid))[1];

  return 0 if (!defined($passwd));

  return (crypt($password, substr($passwd, 0, 2)) eq $passwd);
}

my(@check_data_v4) =
(
  0x0A, 0x5B, 0x31, 0x5D, 0x20, 0x59, 0x6F, 0x75,
  0x20, 0x63, 0x61, 0x6E, 0x20, 0x6D, 0x6F, 0x64,
  0x69, 0x66, 0x79, 0x20, 0x74, 0x68, 0x65, 0x20,
  0x73, 0x6F, 0x75, 0x6E, 0x64, 0x73, 0x20, 0x49,
  0x43, 0x51, 0x20, 0x6D, 0x61, 0x6B, 0x65, 0x73,
  0x2E, 0x20, 0x4A, 0x75, 0x73, 0x74, 0x20, 0x73,
  0x65, 0x6C, 0x65, 0x63, 0x74, 0x20, 0x22, 0x53,
  0x6F, 0x75, 0x6E, 0x64, 0x73, 0x22, 0x20, 0x66,
  0x72, 0x6F, 0x6D, 0x20, 0x74, 0x68, 0x65, 0x20,
  0x22, 0x70, 0x72, 0x65, 0x66, 0x65, 0x72, 0x65,
  0x6E, 0x63, 0x65, 0x73, 0x2F, 0x6D, 0x69, 0x73,
  0x63, 0x22, 0x20, 0x69, 0x6E, 0x20, 0x49, 0x43,
  0x51, 0x20, 0x6F, 0x72, 0x20, 0x66, 0x72, 0x6F,
  0x6D, 0x20, 0x74, 0x68, 0x65, 0x20, 0x22, 0x53,
  0x6F, 0x75, 0x6E, 0x64, 0x73, 0x22, 0x20, 0x69,
  0x6E, 0x20, 0x74, 0x68, 0x65, 0x20, 0x63, 0x6F,
  0x6E, 0x74, 0x72, 0x6F, 0x6C, 0x20, 0x70, 0x61,
  0x6E, 0x65, 0x6C, 0x2E, 0x20, 0x43, 0x72, 0x65,
  0x64, 0x69, 0x74, 0x3A, 0x20, 0x45, 0x72, 0x61,
  0x6E, 0x0A, 0x5B, 0x32, 0x5D, 0x20, 0x43, 0x61,
  0x6E, 0x27, 0x74, 0x20, 0x72, 0x65, 0x6D, 0x65,
  0x6D, 0x62, 0x65, 0x72, 0x20, 0x77, 0x68, 0x61,
  0x74, 0x20, 0x77, 0x61, 0x73, 0x20, 0x73, 0x61,
  0x69, 0x64, 0x3F, 0x20, 0x20, 0x44, 0x6F, 0x75,
  0x62, 0x6C, 0x65, 0x2D, 0x63, 0x6C, 0x69, 0x63,
  0x6B, 0x20, 0x6F, 0x6E, 0x20, 0x61, 0x20, 0x75,
  0x73, 0x65, 0x72, 0x20, 0x74, 0x6F, 0x20, 0x67,
  0x65, 0x74, 0x20, 0x61, 0x20, 0x64, 0x69, 0x61,
  0x6C, 0x6F, 0x67, 0x20, 0x6F, 0x66, 0x20, 0x61,
  0x6C, 0x6C, 0x20, 0x6D, 0x65, 0x73, 0x73, 0x61,
  0x67, 0x65, 0x73, 0x20, 0x73, 0x65, 0x6E, 0x74,
  0x20, 0x69, 0x6E, 0x63, 0x6F, 0x6D, 0x69, 0x6E
);

my(@check_data_v5) =
(
  0x59, 0x60, 0x37, 0x6B, 0x65, 0x62, 0x46, 0x48,
  0x53, 0x61, 0x4C, 0x59, 0x60, 0x57, 0x5B, 0x3D,   # Y`7kebFHSaLY`W[=
  0x5E, 0x34, 0x6D, 0x36, 0x50, 0x3F, 0x6F, 0x67,
  0x53, 0x61, 0x4C, 0x59, 0x40, 0x47, 0x63, 0x39,   # ^4m6P?ogSaLY@Gc9
  0x50, 0x5F, 0x5F, 0x3F, 0x6F, 0x47, 0x43, 0x69,
  0x48, 0x33, 0x31, 0x64, 0x35, 0x5A, 0x4A, 0x42,   # P__?oGCiH31d5ZJB
  0x56, 0x40, 0x67, 0x53, 0x41, 0x07, 0x6C, 0x49,
  0x58, 0x3B, 0x4D, 0x46, 0x68, 0x43, 0x69, 0x48,   # V@gSA.lIX;MFhCiH
  0x33, 0x31, 0x44, 0x65, 0x62, 0x46, 0x48, 0x53,
  0x41, 0x07, 0x6C, 0x69, 0x48, 0x33, 0x51, 0x54,   # 31DebFHSA.liH3QT
  0x5D, 0x4E, 0x6C, 0x49, 0x38, 0x4B, 0x55, 0x4A,
  0x62, 0x46, 0x48, 0x33, 0x51, 0x34, 0x6D, 0x36,   # ]NlI8KUJbFH3Q4m6
  0x50, 0x5F, 0x5F, 0x5F, 0x3F, 0x6F, 0x47, 0x63,
  0x59, 0x40, 0x67, 0x33, 0x31, 0x64, 0x35, 0x5A,   # P___?oGcY@g31d5Z
  0x6A, 0x52, 0x6E, 0x3C, 0x51, 0x34, 0x6D, 0x36,
  0x50, 0x5F, 0x5F, 0x3F, 0x4F, 0x37, 0x4B, 0x35,   # jRn<Q4m6P__?O7K5
  0x5A, 0x4A, 0x62, 0x66, 0x58, 0x3B, 0x4D, 0x66,
  0x58, 0x5B, 0x5D, 0x4E, 0x6C, 0x49, 0x58, 0x3B,   # ZJbfX;MfX[]NlIX;
  0x4D, 0x66, 0x58, 0x3B, 0x4D, 0x46, 0x48, 0x53,
  0x61, 0x4C, 0x59, 0x40, 0x67, 0x33, 0x31, 0x64,   # MfX;MFHSaLY@g31d
  0x55, 0x6A, 0x32, 0x3E, 0x44, 0x45, 0x52, 0x6E,
  0x3C, 0x31, 0x64, 0x55, 0x6A, 0x52, 0x4E, 0x6C,   # Uj2>DERn<1dUjRNl
  0x69, 0x48, 0x53, 0x61, 0x4C, 0x39, 0x30, 0x6F,
  0x47, 0x63, 0x59, 0x60, 0x57, 0x5B, 0x3D, 0x3E,   # iHSaL90oGcY`W[=>
  0x64, 0x35, 0x3A, 0x3A, 0x5A, 0x6A, 0x52, 0x4E,
  0x6C, 0x69, 0x48, 0x53, 0x61, 0x6C, 0x49, 0x58,   # d5::ZjRNliHSalIX
  0x3B, 0x4D, 0x46, 0x68, 0x63, 0x39, 0x50, 0x5F,
  0x5F, 0x3F, 0x6F, 0x67, 0x53, 0x41, 0x25, 0x41,   # ;MFhc9P__?ogSA%A
  0x3C, 0x51, 0x54, 0x3D, 0x5E, 0x54, 0x5D, 0x4E,
  0x4C, 0x39, 0x50, 0x5F, 0x5F, 0x5F, 0x3F, 0x6F,   # <QT=^T]NL9P___?o
  0x47, 0x43, 0x69, 0x48, 0x33, 0x51, 0x54, 0x5D,
  0x6E, 0x3C, 0x31, 0x64, 0x35, 0x5A, 0x00, 0x00,   # GCiH3QT]n<1d5Z..
);

sub decrypt_packet_v4
{
  my $self   = shift;
  my $packet = shift;

  return $packet;
}

sub crypt_packet_v5
{
  my $self      = shift;
  my $packet    = shift;
  my $checkcode = shift;

  use integer;
  
  my($packet_length) = length($packet);

  my($code1) = $packet_length * 0x68656C6C; # 0x68656C6C == 'hell'
  my($code2) = $code1 + $checkcode;
  my($pos) = 0;

#  print("Code1:     ", udword_to_hex($code1), "\n",
#       "Code2:     ", udword_to_hex($code2), "\n",
#       "Checkcode: ", udword_to_hex($checkcode), "\n");

  my($result) = substr($packet, 0, 10);
  $packet = substr($packet, 10) . "\0\0\0\0";
  # Make sure the packet is x*4 bytes long

  my($n) = $packet_length - 10;
  my($data);
  while ($pos < $n)
  {
    my($t)     = ($pos+10) % 256;
    my($code3) = $code2 + $check_data_v5[$t];

#    print("Code3:     ", udword_to_hex($code3), "\n");

    ($packet, $data) = extract_udword($packet);
    $data ^= $code3;

    $result .= pack("V", $data);
    
    $pos += 4;
  }

  # The checkcode wasn't encrypted, so reinsert it
  $result = $self->insert_checkcode_v5($result, $checkcode);
  
  if (length($result) > $packet_length)
  {
    $result = substr($result, 0, $packet_length);
  }
  
  return $result;
}

sub find_checkcode_v5
{
  my $self   = shift;
  my $packet = shift;

  my($code) = unpack("V", substr($packet, 20, 4));
  my($a1, $a2, $a3, $a4, $a5) =
  (
    $code & 0x0001F000,
    $code & 0x07C007C0,
    $code & 0x003E0001,
    $code & 0xF8000000,
    $code & 0x0000083E,
  );

  $a1 >>= 12;
  $a2 >>= 1;
  $a3 <<= 10;
  $a4 >>= 16;
  $a5 <<= 15;

  return $a1 + $a2 + $a3 + $a4 + $a5;
}

sub calculate_checkcode_v5
{
  my $self   = shift;
  my $packet = shift;

  my($b2) = unpack("C", substr($packet, 2, 1));
  my($b4) = unpack("C", substr($packet, 4, 1));
  my($b6) = unpack("C", substr($packet, 6, 1));
  my($b8) = unpack("C", substr($packet, 8, 1));

  my($number1) = ($b8 << 24) + ($b4 << 16) + ($b2 << 8) + $b6;

  my($packet_length) = length($packet);
  my($r1) = 0;     # random % ($packet_length - 0x18)
  my($r2) = 100;    # random % 256

  my($x4) = $r1;
  my($x3) = ~unpack("C", substr($packet, $x4, 1)) & 0xFF;
  my($x2) = $r2;
  my($x1) = ~$check_data_v5[$x2] & 0xFF;

  my($number2) = ($x4<< 24) + ($x3<< 16) + ($x2 << 8) + $x1;

  return $number1 ^ $number2;
}

sub insert_checkcode_v5
{
  my $self      = shift;
  my $packet    = shift;
  my $checkcode = shift;

  my($a1, $a2, $a3, $a4, $a5) =
  (
    $checkcode & 0x0000001F,
    $checkcode & 0x03E003E0,
    $checkcode & 0xF8000400,
    $checkcode & 0x0000F800,
    $checkcode & 0x041F0000,
  );

  $a1 <<= 12;
  $a2 <<= 1;
  $a3 >>= 10;
  $a4 <<= 16;
  $a5 >>= 15;

  my($code) = $a1 + $a2 + $a3 + $a4 + $a5;

  return
    substr($packet, 0, 20) .
      unpack("a4", pack("V", $code)) .
        substr($packet, 24); 
}

sub insert_checkcode_v3_srv
{
  my $self      = shift;
  my $packet    = shift;
  my $checkcode = shift;

  return
    substr($packet, 0, 12) .
      unpack("a4", pack("V", $checkcode)) .
        substr($packet, 16); 
}

sub insert_checkcode_v5_srv
{
  my $self      = shift;
  my $packet    = shift;
  my $checkcode = shift;

  return
    substr($packet, 0, 17) .
      unpack("a4", pack("V", $checkcode)) .
        substr($packet, 21); 
}

sub insert_byte
{
  my $packet = shift;
  my $byte   = shift;

  return $packet . pack("C", $byte);
}

sub insert_uword
{
  my $packet = shift;
  my $uword  = shift;

  return $packet . pack("v", $uword);
}

sub insert_udword
{
  my $packet = shift;
  my $udword = shift;

  return $packet . pack("V", $udword);
}

sub insert_ipaddr
{
  my $packet = shift;
  my $udword = shift;

  return $packet . pack("N", $udword);
}

sub insert_string
{
  my $packet = shift;
  my $string = shift;

  return $packet . pack("va*", length($string)+1, $string . "\0");
}

sub extract_byte
{
  my $packet = shift;

  my $byte = unpack("C", $packet);

  if (length($packet) > 1)
  {
    $packet = substr($packet, 1);
  }
  else
  {
    $packet = "";
  }

  return ($packet, $byte);
}

sub extract_uword
{
  my $packet = shift;

  my $uword = unpack("v", $packet);

  if (length($packet) > 2)
  {
    $packet = substr($packet, 2);
  }
  else
  {
    $packet = "";
  }

  return ($packet, $uword);
}

sub extract_udword
{
  my $packet = shift;

  my $udword = unpack("V", $packet);

  if (length($packet) > 4)
  {
    $packet = substr($packet, 4);
  }
  else
  {
    $packet = "";
  }

  return ($packet, $udword);
}

sub extract_or_get_user_uin
{
  my $self   = shift;
  my $packet = shift;
  my $user_uin;
  
  if ($self->{'packet_header'}->{'version'} == 2)
  {
    ($packet, $user_uin) = extract_udword($packet);
  }
  elsif ($self->{'packet_header'}->{'version'} == 5)
  {
    $user_uin = $self->{'packet_header'}->{'user_uin'};
  }
  else
  {
    die "Unknown packet version";
  }
  
  return ($packet, $user_uin);
}

sub extract_ipaddr
{
  my $packet = shift;

  my $ip_address = unpack("N", $packet);

  if (length($packet) > 4)
  {
    $packet = substr($packet, 4);
  }
  else
  {
    $packet = "";
  }

  return ($packet, $ip_address);
}

sub extract_string
{
  my $packet = shift;

  my $length = unpack("v", $packet);
  my $string;

  if (length($packet) > 2)
  {
    $packet = substr($packet, 2);
  }
  else
  {
    $packet = "";
  }

  if ($length)
  {
    $string = unpack("A$length", $packet);

    if (length($packet) > $length)
    {
      $packet = substr($packet, $length);
    }
    else
    {
      $packet = "";
    }
  }

  return ($packet, $string);
}

=pod

=item dword_to_chars ( DWORD )

Returns the passed DWORD converted to Intel endian character sequence.

=cut

sub dword_to_chars
{
  my $num = shift;
  my @buf;
  my $buf;

  $buf[0] = ($num>>24) & 0x000000FF;
  $buf[1] = ($num>>16) & 0x000000FF;
  $buf[2] = ($num>>8) & 0x000000FF;
  $buf[3] = ($num) & 0x000000FF;

  $buf  = $buf[3];
  $buf <<= 8;
  $buf |= $buf[2];
  $buf <<= 8;
  $buf |= $buf[1];
  $buf <<= 8;
  $buf |= $buf[0];

  return($buf);
}

=pod

=item uword_to_hex ( NUMBER )

Returns the passed NUMBER in hex representation as a string.

=cut

sub uword_to_hex
{
  return "0x" . uc(sprintf("%04x", $_[0]));
}



=pod

=item udword_to_hex ( NUMBER )

Returns the passed NUMBER in hex representation as a string.

=cut

sub udword_to_hex
{
  return "0x" . uc(sprintf("%08x", $_[0]));
}

=pod

=item min ( A, B )

Returns the smallest of A and B.

=cut

sub min
{
  return $_[0] < $_[1] ? $_[0] : $_[1];
}

sub max
{
  return $_[0] > $_[1] ? $_[0] : $_[1];
}

sub dump_packet
{
  my $self   = shift;
  my $packet = shift;

  my($r) = 1;
  my($size) = length($packet);
  my($buffer);
  my($char);
  for (my($i, $x) = (0, 0); ; $i++)
  {
    if ($i < $size) {
      printf("%04X: ", $i) if(!$x);
      $char = chr(unpack("C", substr($packet, $i, 1)));
      printf("%02X ", ord($char));
      $char = '.' if ($char lt ' ' ||
                      (ord($char) >= 128 && ord($char) < 160));
    }
    else
    {
      last if (!$x);

      printf("   ");
      $char = ' ';
      $r = 0;
    }
    $buffer .= $char;
    $x++;
    if(!($x < 16))
    {
      print "$buffer\n";
      $buffer = '';
      $x = 0;
      last if (!$r);
    }
  }
}

=pod

=head1 AUTHOR

Peter Kjellerstedt <pkj@axis.com> and
Fredrik Norrman <fredrikn@axis.com>

=head1 CREDITS

Bek Oberin <gossamer@tertius.net.au> - for writing Net::ICQ, which
this code was originally based on.

=head1 COPYRIGHT

Copyright (c) 1999 Peter Kjellerstedt and Fredrik Norrman.
All rights reserved.

This program is free software; you can redistribute it and/or modify
it under the same terms as Perl itself (GPL).

=cut

#
# End code.
#
1;
